from typing import Literal, Optional
from dlt.common.configuration import BasicConfiguration

TPoolType = Literal["process", "thread", "none"]

class PoolRunnerConfiguration(BasicConfiguration):
    MAX_PARALLELISM: Optional[int] = None  # how many threads/processes in the pool
    EXIT_ON_EXCEPTION: bool = False  # should exit on exception
    STOP_AFTER_RUNS: int = 10000  # will stop runner with exit code -2 after so many runs, that prevents memory fragmentation
    POOL_TYPE: TPoolType = None  # type of pool to run, must be set in derived configs
    RUN_SLEEP: float = 0.5  # how long to sleep between runs with workload, seconds
    RUN_SLEEP_IDLE: float = 1.0  # how long to sleep when no more items are pending, seconds
    RUN_SLEEP_WHEN_FAILED: float = 1.0  # how long to sleep between the runs when failed



import os
import psycopg2
from psycopg2.sql import SQL, Identifier, Composed, Literal as SQLLiteral
from typing import Any, AnyStr, Dict, List, Literal, Optional, Tuple, Type

from dlt.common.typing import StrAny
from dlt.common.arithmetics import DEFAULT_NUMERIC_PRECISION, DEFAULT_NUMERIC_SCALE
from dlt.common.configuration import PostgresConfiguration
from dlt.common.dataset_writers import TWriterType, escape_redshift_identifier
from dlt.common.schema import COLUMN_HINTS, Column, ColumnBase, DataType, HintType, Schema, SchemaUpdate, Table

from dlt.loaders.exceptions import (LoadClientSchemaWillNotUpdate, LoadClientTerminalInnerException,
                                            LoadClientTransientInnerException, LoadFileTooBig)
from dlt.loaders.local_types import LoadJobStatus
from dlt.loaders.client_base import ClientBase, SqlClientBase, LoadJob

SCT_TO_PGT: Dict[DataType, str] = {
    "text": "varchar(max)",
    "double": "double precision",
    "bool": "boolean",
    "timestamp": "timestamp with time zone",
    "bigint": "bigint",
    "binary": "varbinary",
    "decimal": f"numeric({DEFAULT_NUMERIC_PRECISION},{DEFAULT_NUMERIC_SCALE})"
}

PGT_TO_SCT: Dict[str, DataType] = {
    "varchar(max)": "text",
    "double precision": "double",
    "boolean": "bool",
    "timestamp with time zone": "timestamp",
    "bigint": "bigint",
    "binary varying": "binary",
    "numeric": "decimal"
}

HINT_TO_REDSHIFT_ATTR: Dict[HintType, str] = {
    "cluster": "DISTKEY",
    # it is better to not enforce constraints in redshift
    # "primary_key": "PRIMARY KEY",
    "sort": "SORTKEY"
}


class SqlClientMixin:

    MAX_STATEMENT_SIZE = 16 * 1024 * 1204

    def __init__(self, CONFIG: Type[PostgresConfiguration], *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._conn: psycopg2.connection = None
        self.C = CONFIG

    def _open_connection(self) -> None:
        self._conn = psycopg2.connect(dbname=self.C.PG_DATABASE_NAME,
                             user=self.C.PG_USER,
                             host=self.C.PG_HOST,
                             port=self.C.PG_PORT,
                             password=self.C.PG_PASSWORD,
                             connect_timeout=self.C.PG_CONNECTION_TIMEOUT
                             )
        # we'll provide explicit transactions
        self._conn.set_session(autocommit=True)

    def _close_connection(self) -> None:
        if self._conn:
            self._conn.close()
            self._conn = None

    def _execute_sql(self, query: AnyStr) -> Any:
        curr: psycopg2.cursor
        with self._conn.cursor() as curr:
            try:
                curr.execute(query)
            except psycopg2.Error as outer:
                try:
                    self._conn.rollback()
                    self._conn.reset()
                except psycopg2.Error:
                    self._close_connection()
                    self._open_connection()
                raise outer
            if curr.description is None:
                return None
            else:
                f = curr.fetchall()
                return f


class RedshiftInsertLoadJob(SqlClientMixin, LoadJob):
    def __init__(self, canonical_table_name: str, file_path: str, conn: Any, CONFIG: Type[PostgresConfiguration]) -> None:
        super().__init__(CONFIG, ClientBase.get_file_name_from_file_path(file_path))
        self._conn = conn
        # insert file content immediately
        self._insert(canonical_table_name, file_path)

    def status(self) -> LoadJobStatus:
        # this job is always done
        return "completed"

    def file_name(self) -> str:
        return self._file_name

    def exception(self) -> str:
        # this part of code should be never reached
        raise NotImplementedError()

    def _insert(self, canonical_table_name: str, file_path: str) -> None:
        # TODO: implement tracking of jobs in storage, both completed and failed
        # WARNING: maximum redshift statement is 16MB https://docs.aws.amazon.com/redshift/latest/dg/c_redshift-sql.html
        # in case of postgres: 2GiB
        if os.stat(file_path).st_size >= SqlClientMixin.MAX_STATEMENT_SIZE:
            # terminal exception
            raise LoadFileTooBig(file_path, SqlClientMixin.MAX_STATEMENT_SIZE)
        with open(file_path, "r") as f:
            header = f.readline()
            content = f.read()
        sql = Composed(
            [SQL("BEGIN TRANSACTION;"),
            SQL(header).format(SQL(canonical_table_name)),
            SQL(content),
            SQL("COMMIT TRANSACTION;")]
        )
        self._execute_sql(sql)


class RedshiftClient(SqlClientMixin, SqlClientBase):
    def __init__(self, schema: Schema, CONFIG: Type[PostgresConfiguration]) -> None:
        super().__init__(CONFIG, schema)

    def initialize_storage(self) -> None:
        schema_name = self._to_canonical_schema_name()
        query = """
                SELECT 1
                    FROM INFORMATION_SCHEMA.SCHEMATA
                    WHERE schema_name = {};
                """
        rows = self._execute_sql(SQL(query).format(SQLLiteral(schema_name)))
        if len(rows) == 0:
            self._execute_sql(SQL("CREATE SCHEMA {};").format(Identifier(schema_name)))

    def get_file_load(self, file_path: str) -> LoadJob:
        # always returns completed jobs as RedshiftInsertLoadJob is executed
        # atomically in start_file_load so any jobs that should be recreated are already completed
        # in case of bugs in loader (asking for jobs that were never created) we are not able to detect that
        return ClientBase.make_job_with_status(file_path, "completed")

    def start_file_load(self, table_name: str, file_path: str) -> LoadJob:
        # verify that table exists in the schema
        self._get_table_by_name(table_name, file_path)
        try:
            return RedshiftInsertLoadJob(self._to_canonical_table_name(table_name), file_path, self._conn, self.C)
        except (psycopg2.OperationalError, psycopg2.InternalError) as tr_ex:
            if tr_ex.pgerror is not None:
                if "Cannot insert a NULL value into column" in tr_ex.pgerror:
                    # NULL violations is internal error, probably a redshift thing
                    raise LoadClientTerminalInnerException("Terminal error, file will not load", tr_ex)
                if "Numeric data overflow" in tr_ex.pgerror:
                    raise LoadClientTerminalInnerException("Terminal error, file will not load", tr_ex)
                if "Precision exceeds maximum":
                    raise LoadClientTerminalInnerException("Terminal error, file will not load", tr_ex)
            raise LoadClientTransientInnerException("Error may go away, will retry", tr_ex)
        except (psycopg2.DataError, psycopg2.ProgrammingError, psycopg2.IntegrityError) as ter_ex:
            raise LoadClientTerminalInnerException("Terminal error, file will not load", ter_ex)

    def update_storage_schema(self) -> None:
        storage_version = self._get_schema_version_from_storage()
        if storage_version < self.schema.schema_version:
            for sql in self._build_schema_update_sql():
                self._execute_sql(sql)
            self._update_schema_version(self.schema.schema_version)

    def _get_schema_version_from_storage(self) -> int:
        try:
            return super()._get_schema_version_from_storage()
        except psycopg2.ProgrammingError:
            # there's no table so there's no schema
            return 0

    def _build_schema_update_sql(self) -> List[str]:
        sql_updates = []
        for table_name in self.schema.schema_tables:
            exists, storage_table = self._get_storage_table(table_name)
            sql = self._get_table_update_sql(table_name, storage_table, exists)
            if sql:
                sql_updates.append(sql)
        return sql_updates

    def _get_table_update_sql(self, table_name: str, storage_table: Table, exists: bool) -> str:
        new_columns = self._create_table_update(table_name, storage_table)
        if len(new_columns) == 0:
            # no changes
            return None
        # build sql
        canonical_name = self._to_canonical_table_name(table_name)
        sql = "BEGIN TRANSACTION;\n"
        if not exists:
            # build CREATE
            sql += f"CREATE TABLE {canonical_name} (\n"
            sql += ",\n".join([self._get_column_def_sql(c) for c in new_columns])
            sql += ");"
        else:
            # build ALTER as separate statement for each column (redshift limitation)
            sql += "\n".join([f"ALTER TABLE {canonical_name}\nADD COLUMN {self._get_column_def_sql(c)};" for c in new_columns])
        # scan columns to get hints
        if exists:
            # no hints may be specified on added columns
            for hint in COLUMN_HINTS:
                if any(c.get(hint, False) is True for c in new_columns):
                    hint_columns = [c["name"] for c in new_columns if c.get(hint, False)]
                    raise LoadClientSchemaWillNotUpdate(canonical_name, hint_columns, f"{hint} requested after table was created")
        # TODO: add FK relations
        sql += "\nCOMMIT TRANSACTION;"
        return sql

    def _get_column_def_sql(self, c: Column) -> str:
        hints_str = " ".join(HINT_TO_REDSHIFT_ATTR.get(h, "") for h in HINT_TO_REDSHIFT_ATTR.keys() if c.get(h, False) is True)
        column_name = escape_redshift_identifier(c["name"])
        return f"{column_name} {self._sc_t_to_pq_t(c['data_type'])} {hints_str} {self._gen_not_null(c['nullable'])}"

    def _get_storage_table(self, table_name: str) -> Tuple[bool, Table]:
        schema_table: Table = {}
        query = f"""
                SELECT column_name, data_type, is_nullable, numeric_precision, numeric_scale
                    FROM INFORMATION_SCHEMA.COLUMNS
                WHERE table_schema = '{self._to_canonical_schema_name()}' AND table_name = '{table_name}'
                ORDER BY ordinal_position;
                """
        rows = self._execute_sql(query)
        # if no rows we assume that table does not exist
        if len(rows) == 0:
            # TODO: additionally check if table exists
            return False, schema_table
        # TODO: pull more data to infer DISTKEY, PK and SORTKEY attributes/constraints
        for c in rows:
            schema_c: ColumnBase = {
                "name": c[0],
                "nullable": self._null_to_bool(c[2]),
                "data_type": self._pq_t_to_sc_t(c[1], c[3], c[4]),
            }
            schema_table[c[0]] = Schema._add_missing_hints(schema_c)
        return True, schema_table


    def _to_canonical_schema_name(self) -> str:
        return f"{self.C.PG_SCHEMA_PREFIX}_{self.schema.schema_name}"

    def _to_canonical_table_name(self, table_name: str) -> str:
        return f"{self._to_canonical_schema_name()}.{table_name}"

    @staticmethod
    def _null_to_bool(v: str) -> bool:
        if v == "NO":
            return False
        elif v == "YES":
            return True;
        raise ValueError(v)

    @staticmethod
    def _gen_not_null(v: bool) -> str:
        return "NOT NULL" if not v else ""

    @staticmethod
    def _sc_t_to_pq_t(sc_t: DataType) -> str:
        if sc_t == "wei":
            return f"numeric({DEFAULT_NUMERIC_PRECISION},0)"
        return SCT_TO_PGT[sc_t]

    @staticmethod
    def _pq_t_to_sc_t(pq_t: str, precision: Optional[int], scale: Optional[int]) -> DataType:
        if pq_t == "numeric":
            if precision == DEFAULT_NUMERIC_PRECISION and scale == 0:
                return "wei"
        return PGT_TO_SCT.get(pq_t, "text")


def make_client(schema: Schema, C: Type[PostgresConfiguration]) -> RedshiftClient:
    return RedshiftClient(schema, C)


def supported_writer(C: Type[PostgresConfiguration]) -> TWriterType:
    return "insert_values"



from typing import Sequence
from dlt.common.exceptions import DltException, TerminalException, TransientException

from dlt.loaders.local_types import LoadJobStatus


class LoadException(DltException):
    def __init__(self, msg: str) -> None:
        super().__init__(msg)


class LoadClientTerminalException(LoadException, TerminalException):
    def __init__(self, msg: str) -> None:
        super().__init__(msg)


class LoadClientTransientException(LoadException, TransientException):
    def __init__(self, msg: str) -> None:
        super().__init__(msg)


class LoadClientTerminalInnerException(LoadClientTerminalException):
    def __init__(self, msg: str, inner_exc: Exception) -> None:
        self.inner_exc = inner_exc
        super().__init__(msg)


class LoadClientTransientInnerException(LoadClientTransientException):
    def __init__(self, msg: str, inner_exc: Exception) -> None:
        self.inner_exc = inner_exc
        super().__init__(msg)



class LoadJobNotExistsException(LoadClientTerminalException):
    def __init__(self, job_id: str) -> None:
        super().__init__(f"Job with id/file name {job_id} not found")


class LoadUnknownTableException(LoadClientTerminalException):
    def __init__(self, table_name: str, file_name: str) -> None:
        self.table_name = table_name
        super().__init__(f"Client does not know table {table_name} for load file {file_name}")


class LoadJobInvalidStateTransitionException(LoadClientTerminalException):
    def __init__(self, from_state: LoadJobStatus, to_state: LoadJobStatus) -> None:
        self.from_state = from_state
        self.to_state = to_state
        super().__init__(f"Load job cannot transition form {from_state} to {to_state}")

class LoadJobServerTerminalException(LoadClientTerminalException):
    def __init__(self, file_path: str) -> None:
        super().__init__(f"Job with id/file name {file_path} encountered unrecoverable problem")


class LoadClientSchemaVersionCorrupted(LoadClientTerminalException):
    def __init__(self, dataset_name: str) -> None:
        self.dataset_name = dataset_name
        super().__init__(f"Schema _version table contains too many rows in {dataset_name}")


class LoadClientSchemaWillNotUpdate(LoadClientTerminalException):
    def __init__(self, table_name: str, columns: Sequence[str], msg: str) -> None:
        self.table_name = table_name
        self.columns = columns
        super().__init__(f"Schema for table {table_name} column(s) {columns} will not update: {msg}")


class LoadFileTooBig(LoadClientTerminalException):
    def __init__(self, file_name: str, max_size: int) -> None:
        super().__init__(f"File {file_name} exceedes {max_size} and cannot be loaded. Split the file and try again.")



import argparse
import multiprocessing
from prometheus_client import Counter, Gauge, Summary, CollectorRegistry, REGISTRY
from typing import Callable, Dict, NamedTuple, Optional, Type, TypeVar, Union, cast
from multiprocessing.pool import ThreadPool, Pool

from dlt.common import logger, signals
from dlt.common.configuration.basic_configuration import BasicConfiguration
from dlt.common.time import sleep
from dlt.common.telemetry import TRunHealth, TRunMetrics, get_logging_extras, get_metrics_from_prometheus
from dlt.common.logger import init_logging_from_config, init_telemetry, process_internal_exception
from dlt.common.signals import register_signals
from dlt.common.utils import str2bool
from dlt.common.exceptions import SignalReceivedException, TimeRangeExhaustedException, UnsupportedProcessStartMethodException
from dlt.common.configuration import PoolRunnerConfiguration

TPool = TypeVar("TPool", bound=Pool)


class TRunArgs(NamedTuple):
    single_run: bool
    wait_runs: int


RUN_ARGS = TRunArgs(False, 0)

HEALTH_PROPS_GAUGES: Dict[str, Union[Counter, Gauge]] = None
RUN_DURATION_GAUGE: Gauge = None
RUN_DURATION_SUMMARY: Summary = None

LAST_RUN_METRICS: TRunMetrics = None
LAST_RUN_EXCEPTION: BaseException = None

def create_gauges(registry: CollectorRegistry) -> None:
    global HEALTH_PROPS_GAUGES, RUN_DURATION_GAUGE, RUN_DURATION_SUMMARY

    HEALTH_PROPS_GAUGES = {
        "runs_count": Counter("runs_count", "Count runs", registry=registry),
        "runs_not_idle_count": Counter("runs_not_idle_count", "Count not idle runs", registry=registry),
        "runs_healthy_count": Counter("runs_healthy_count", "Count healthy runs", registry=registry),
        "runs_cs_healthy_gauge": Gauge("runs_cs_healthy_gauge", "Count consecutive healthy runs, reset on failed run", registry=registry),
        "runs_failed_count": Counter("runs_failed_count", "Count failed runs", registry=registry),
        "runs_cs_failed_gauge": Gauge("runs_cs_failed_gauge", "Count consecutive failed runs, reset on healthy run", registry=registry),
        "runs_pending_items_gauge": Gauge("runs_pending_items_gauge", "Number of items pending at the end of the run", registry=registry),
    }

    RUN_DURATION_GAUGE = Gauge("runs_duration_seconds", "Duration of the run", registry=registry)
    RUN_DURATION_SUMMARY = Summary("runs_duration_summary", "Summary of the run duration", registry=registry)


def update_gauges() -> TRunHealth:
    return get_metrics_from_prometheus(HEALTH_PROPS_GAUGES.values())  # type: ignore


def str2bool_a(v: str) -> bool:
    try:
        return str2bool(v)
    except ValueError:
        raise argparse.ArgumentTypeError('Boolean value expected.')


def create_default_args(C: Type[PoolRunnerConfiguration]) -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description=f"Default runner for {C.NAME}")
    add_pool_cli_arguments(parser)
    return parser


def add_pool_cli_arguments(parser: argparse.ArgumentParser) -> None:
    parser.add_argument("--single-run", type=str2bool_a, nargs='?', const=True, default=False, help="exit when all pending items are processed")
    parser.add_argument("--wait-runs", type=int, nargs='?', const=True, default=1, help="maximum idle runs to wait for incoming data")



def initialize_runner(C: Type[BasicConfiguration], run_args: Optional[TRunArgs] = None) -> None:
    global RUN_ARGS

    if run_args is not None:
        RUN_ARGS = run_args

    # initialize or re-initialize logging with new settings
    init_logging_from_config(C)

    # initialize only once
    if not HEALTH_PROPS_GAUGES:
        init_telemetry(C)
        create_gauges(REGISTRY)
        register_signals()



def pool_runner(C: Type[PoolRunnerConfiguration], run_f: Callable[[TPool], TRunMetrics]) -> int:
    # start pool
    pool: Pool = None
    if C.POOL_TYPE == "process":
        # our pool implementation do not work on spawn
        if multiprocessing.get_start_method() != "fork":
            raise UnsupportedProcessStartMethodException(multiprocessing.get_start_method())
        pool = Pool(processes=C.MAX_PARALLELISM)
    elif C.POOL_TYPE == "thread":
        pool = ThreadPool(processes=C.MAX_PARALLELISM)
    else:
        pool = None
    logger.info(f"Created {C.POOL_TYPE} pool with {C.MAX_PARALLELISM or 'default no.'} workers")


    try:
        while True:
            run_metrics: TRunMetrics = None
            try:
                HEALTH_PROPS_GAUGES["runs_count"].inc()
                # run pool logic
                with RUN_DURATION_SUMMARY.time(), RUN_DURATION_GAUGE.time():
                    run_metrics = run_f(cast(TPool, pool))
            except Exception as exc:
                if (type(exc) is SignalReceivedException) or (type(exc) is TimeRangeExhaustedException):
                    # always exit
                    raise
                else:
                    process_internal_exception("run")
                    # the run failed
                    run_metrics = TRunMetrics(True, True, -1)
                    # preserve exception
                    global LAST_RUN_EXCEPTION
                    LAST_RUN_EXCEPTION = exc

            # gather and emit metrics
            if not run_metrics.was_idle:
                HEALTH_PROPS_GAUGES["runs_not_idle_count"].inc()
            if run_metrics.has_failed:
                HEALTH_PROPS_GAUGES["runs_failed_count"].inc()
                HEALTH_PROPS_GAUGES["runs_cs_failed_gauge"].inc()
                HEALTH_PROPS_GAUGES["runs_cs_healthy_gauge"].set(0)
            else:
                HEALTH_PROPS_GAUGES["runs_healthy_count"].inc()
                HEALTH_PROPS_GAUGES["runs_cs_healthy_gauge"].inc()
                HEALTH_PROPS_GAUGES["runs_cs_failed_gauge"].set(0)
            HEALTH_PROPS_GAUGES["runs_pending_items_gauge"].set(run_metrics.pending_items)
            health_props = update_gauges()
            logger.health("run health counters", extra={"metrics": health_props})
            logger.metrics("run metrics", extra=get_logging_extras([RUN_DURATION_GAUGE, RUN_DURATION_SUMMARY]))

            # preserve last run metrics
            global LAST_RUN_METRICS
            LAST_RUN_METRICS = run_metrics

            # exit due to signal
            signals.raise_if_signalled()

            # exit due to exception and flag
            if run_metrics.has_failed and C.EXIT_ON_EXCEPTION:
                logger.warning(f"Exiting runner due to EXIT_ON_EXCEPTION flag set")
                return -1

            # single run may be forced but at least wait_runs must pass
            if RUN_ARGS.single_run and (health_props["runs_count"] >= RUN_ARGS.wait_runs and
                # and was all the time idle or (was not idle but now pending is 0)
                (health_props["runs_not_idle_count"] == 0 or run_metrics.pending_items == 0)):
                logger.warning(f"Stopping runner due to single run override")
                return 0

            if run_metrics.has_failed:
                sleep(C.RUN_SLEEP_WHEN_FAILED)
            elif run_metrics.pending_items == 0:
                # nothing is pending so we can sleep longer
                sleep(C.RUN_SLEEP_IDLE)
            else:
                # more items are pending, sleep (typically) shorter
                sleep(C.RUN_SLEEP)

            # this allows to recycle long living process that get their memory fragmented
            # exit after runners sleeps so we keep the running period
            if health_props["runs_count"] == C.STOP_AFTER_RUNS:
                logger.warning(f"Stopping runner due to max runs {health_props['runs_count']} exceeded")
                return -2
    except SignalReceivedException as sigex:
        # sleep this may raise SignalReceivedException
        logger.warning(f"Exiting runner due to signal {sigex.signal_code}")
        return sigex.signal_code
    except TimeRangeExhaustedException as tre:
        logger.info(f"{str(tre)}, not further processing will be done")
        return 0
    finally:
        if pool:
            logger.info("Closing processing pool")
            pool.close()
            pool.join()
            pool = None



from dlt.common.dataset_writers import TWriterType

class DummyClientConfiguration:
    WRITER_TYPE: TWriterType = "jsonl"
    FAIL_PROB: float = 0.0
    RETRY_PROB: float = 0.0
    COMPLETED_PROB: float = 0.0
    TIMEOUT: float = 10.0



class UnpackingVolumeConfiguration:
    UNPACKING_VOLUME_PATH: str = "_storage/unpacking"  # path to volume where unpacking will happen


class ProductionUnpackingVolumeConfiguration:
    UNPACKING_VOLUME_PATH: str = None



import hashlib
from os import environ
import secrets
from typing import Any, Iterator, Sequence, TypeVar, Mapping, List, Union

from dlt.common.typing import StrAny, DictStrAny, StrStr

T = TypeVar("T")


def chunks(list: Sequence[T], n: int) -> Iterator[Sequence[T]]:
    for i in range(0, len(list), n):
        yield list[i:i + n]


def uniq_id() -> str:
    return secrets.token_hex(16)


def digest128(v: str) -> str:
    return hashlib.shake_128(v.encode("utf-8")).hexdigest(16)


def str2bool(v: str) -> bool:
    if isinstance(v, bool):
        return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise ValueError('Boolean value expected.')


def flatten_list_of_dicts(dicts: Sequence[StrAny]) -> StrAny:
    """
    Transforms a list of objects [{K: {...}}, {L: {....}}, ...] -> {K: {...}, L: {...}...}
    """
    o: DictStrAny = {}
    for d in dicts:
        for k,v in d.items():
            if k in o:
                raise KeyError(f"Cannot flatten with duplicate key {k}")
            o[k] = v
    return o


def flatten_list_of_str_or_dicts(seq: Sequence[Union[StrAny, str]]) -> StrAny:
    """
    Transforms a list of objects or strings [{K: {...}}, L, ...] -> {K: {...}, L: None, ...}
    """
    o: DictStrAny = {}
    for e in seq:
        if isinstance(e, dict):
            for k,v in e.items():
                if k in o:
                    raise KeyError(f"Cannot flatten with duplicate key {k}")
                o[k] = v
        else:
            key = str(e)
            if key in o:
                raise KeyError(f"Cannot flatten with duplicate key {k}")
            o[key] = None
    return o


def flatten_dicts_of_dicts(dicts: Mapping[str, Any]) -> Sequence[Any]:
    """
    Transform and object {K: {...}, L: {...}...} -> [{key:K, ....}, {key: L, ...}, ...]
    """
    o: List[Any] = []
    for k, v in dicts.items():
        if isinstance(v, list):
            # if v is a list then add "key" to each list element
            for lv in v:
                lv["key"] = k
        else:
            # add as "key" to dict
            v["key"] = k

        o.append(v)
    return o


def tuplify_list_of_dicts(dicts: Sequence[DictStrAny]) -> Sequence[DictStrAny]:
    """
    Transform dicts with single key into {"key": orig_key, "value": orig_value}
    """
    for d in dicts:
        if len(d) > 1:
            raise ValueError(f"Tuplify requires one key dicts {d}")
        if len(d) == 1:
            key = next(iter(d))
            # delete key first to avoid name clashes
            value = d[key]
            del d[key]
            d["key"] = key
            d["value"] = value

    return dicts


def filter_env_vars(vars: List[str]) -> StrStr:
    return {k.lower(): environ[k] for k in vars if k in environ}


def update_dict_with_prune(dest: DictStrAny, update: StrAny) -> None:
    for k, v in update.items():
        if v is not None:
            dest[k] = v
        elif k in dest:
            del dest[k]


def is_interactive() -> bool:
    import __main__ as main
    return not hasattr(main, '__file__')



from typing import Callable, Dict, Any, List, Literal, Mapping, Sequence, TypeVar, TypedDict, Optional, Union

DictStrAny = Dict[str, Any]
DictStrStr = Dict[str, str]
StrAny = Mapping[str, Any]  # immutable, covariant entity
StrStr = Mapping[str, str]  # immutable, covariant entity
StrStrStr = Mapping[str, Mapping[str, str]]  # immutable, covariant entity
TFun = TypeVar("TFun", bound=Callable[..., Any])


class TEvent(TypedDict, total=False):
    pass


class TTimestampEvent(TEvent, total=False):
    timestamp: float  # timestamp of event



import os
import logging
import tempfile
from typing import Any, Iterator, List, Sequence
from git import Repo, Git, RepositoryDirtyError
from contextlib import contextmanager

from dlt.common import json
from dlt.common.utils import uniq_id
from dlt.common.typing import StrAny, Optional
from dlt.dbt_runner.exceptions import DBTRunnerException

# block disabling root logger
import logbook.compat
logbook.compat.redirect_logging = lambda : None

# can only import DBT after redirect is disabled
import dbt.main
import dbt.logger
from dbt.events import functions
from dbt.contracts import results as dbt_results
from dbt.exceptions import FailFastException


# keep this exception definition here due to mock of logbook
class DBTProcessingError(DBTRunnerException):
    def __init__(self, command: str, results: Any) -> None:
        self.command = command
        # the results from DBT may be anything
        self.results = results
        super().__init__(f"DBT command {command} could not be executed")


@contextmanager
def git_custom_key_command(private_key: Optional[str]) -> Iterator[str]:
    if private_key:
        key_file = tempfile.mktemp(prefix=uniq_id())
        with open(key_file, "w") as f:
            f.write(private_key)
        try:
            # permissions so SSH does not complain
            os.chmod(key_file, 0o600)
            yield 'ssh -o "StrictHostKeyChecking accept-new" -i %s' % key_file
        finally:
            os.remove(key_file)
    else:
        yield 'ssh -o "StrictHostKeyChecking accept-new"'


def ensure_remote_head(repo_path: str, with_git_command: Optional[str] = None) -> None:
    # update remotes and check if heads are same. ignores locally modified files
    repo = Repo(repo_path)
    # use custom environemnt if specified
    with repo.git.custom_environment(GIT_SSH_COMMAND=with_git_command):
        # update origin
        repo.remote().update()
        # get branch status
        status: str = repo.git.status("--short", "--branch", "-uno")
        # we expect first status line ## main...origin/main
        status_line = status.split("/n")[0]
        if not (status_line.startswith("##") and not status_line.endswith("]")):
            raise RepositoryDirtyError(repo, status)


def clone_repo(repository_url: str, clone_path: str, branch: Optional[str] = None, with_git_command: Optional[str] = None) -> None:
    repo = Repo.clone_from(repository_url, clone_path, env=dict(GIT_SSH_COMMAND=with_git_command))
    if branch:
        repo.git.checkout(branch)


def initialize_dbt_logging(level: str, is_json_logging: bool) -> Sequence[str]:
    int_level = logging._nameToLevel[level]

    # wrap log setup to force out log level

    def setup_event_logger_wrapper(log_path: str, level_override:str = None) -> None:
        functions.setup_event_logger(log_path, level)
        # force log level as file is debug only
        functions.this.FILE_LOG.setLevel(level)
        functions.this.FILE_LOG.handlers[0].setLevel(level)

    dbt.main.setup_event_logger = setup_event_logger_wrapper

    globs = []
    if int_level <= logging.DEBUG:
        globs = ["--debug"]

    # return global parameters to be passed to setup logging

    if is_json_logging:
        return ["--log-format", "json"] + globs
    else:
        return globs


def is_incremental_schema_out_of_sync_error(error: dbt_results.RunResult) -> bool:
    return issubclass(type(error), dbt_results.RunResult) and error.status == dbt_results.RunStatus.Error and\
        "The source and target schemas on this incremental model are out of sync" in error.message


def run_dbt_command(package_path: str, command: str, profiles_dir: str, profile_name: Optional[str] = None,
                    global_args: Sequence[str] = None, command_args: Sequence[str] = None, vars: StrAny = None) -> Sequence[dbt_results.BaseResult]:
    args = ["--profiles-dir", profiles_dir]
    # add profile name if provided
    if profile_name:
        args += ["--profile", profile_name]
    # serialize dbt variables to pass to package
    if vars:
        args += ["--vars", json.dumps(vars)]
    if command_args:
        args += command_args

    # cwd to package dir
    working_dir = os.getcwd()
    os.chdir(package_path)
    try:
        results: List[dbt_results.BaseResult] = None
        success: bool = None
        results, success = dbt.main.handle_and_check((global_args or []) + [command] + args)  # type: ignore
        assert type(success) is bool
        if not success:
            raise DBTProcessingError(command ,results)
        return results
    except FailFastException as ff:
        raise DBTProcessingError(command, ff.result) from ff
    finally:
        # unblock logger manager to run next command
        dbt.logger.log_manager.reset_handlers()
        # go back to working dir
        os.chdir(working_dir)



from pathlib import Path
from typing import Any, AnyStr, Dict, List, Literal, Optional, Tuple, Type
import google.cloud.bigquery as bigquery
from google.cloud import exceptions as gcp_exceptions
from google.oauth2 import service_account
from google.api_core import exceptions as api_core_exceptions


from dlt.common import json, logger
from dlt.common.arithmetics import DEFAULT_NUMERIC_PRECISION, DEFAULT_NUMERIC_SCALE
from dlt.common.configuration import GcpClientConfiguration
from dlt.common.dataset_writers import TWriterType, escape_bigquery_identifier
from dlt.loaders.local_types import LoadJobStatus
from dlt.common.schema import Column, DataType, Schema, Table

from dlt.loaders.client_base import SqlClientBase, LoadJob
from dlt.loaders.exceptions import LoadClientSchemaWillNotUpdate, LoadJobNotExistsException, LoadJobServerTerminalException, LoadUnknownTableException

SCT_TO_BQT: Dict[DataType, str] = {
    "text": "STRING",
    "double": "FLOAT64",
    "bool": "BOOLEAN",
    "timestamp": "TIMESTAMP",
    "bigint": "INTEGER",
    "binary": "BYTES",
    "decimal": f"NUMERIC({DEFAULT_NUMERIC_PRECISION},{DEFAULT_NUMERIC_SCALE})",
    "wei": "BIGNUMERIC"  # non parametrized should hold wei values
}

BQT_TO_SCT: Dict[str, DataType] = {
    "STRING": "text",
    "FLOAT": "double",
    "BOOLEAN": "bool",
    "TIMESTAMP": "timestamp",
    "INTEGER": "bigint",
    "BYTES": "binary",
    "NUMERIC": "decimal",
    "BIGNUMERIC": "decimal"
}

class BigQueryLoadJob(LoadJob):
    def __init__(self, file_name: str, bq_load_job: bigquery.LoadJob, CONFIG: Type[GcpClientConfiguration]) -> None:
        self.bq_load_job = bq_load_job
        self.C = CONFIG
        self.default_retry = bigquery.DEFAULT_RETRY.with_deadline(CONFIG.TIMEOUT)
        super().__init__(file_name)

    def status(self) -> LoadJobStatus:
        # check server if done
        done = self.bq_load_job.done(retry=self.default_retry, timeout=self.C.TIMEOUT)
        if done:
            # rows processed
            if self.bq_load_job.output_rows is not None and self.bq_load_job.error_result is None:
                return "completed"
            else:
                return "failed"
        else:
            return "running"

    def file_name(self) -> str:
        return self._file_name

    def exception(self) -> str:
        exception: str = json.dumps({
            "error_result": self.bq_load_job.error_result,
            "errors": self.bq_load_job.errors,
            "job_start": self.bq_load_job.started,
            "job_end": self.bq_load_job.ended,
            "job_id": self.bq_load_job.job_id
        })
        return exception


class BigQueryClient(SqlClientBase):
    def __init__(self, schema: Schema, CONFIG: Type[GcpClientConfiguration]) -> None:
        self._client: bigquery.Client = None
        self.C = CONFIG
        self.default_retry = bigquery.DEFAULT_RETRY.with_deadline(CONFIG.TIMEOUT)
        super().__init__(schema)


    def initialize_storage(self) -> None:
        dataset_name = self._to_canonical_schema_name()
        try:
            self._client.get_dataset(dataset_name, retry=self.default_retry, timeout=self.C.TIMEOUT)
        except gcp_exceptions.NotFound:
            self._client.create_dataset(dataset_name, exists_ok=False, retry=self.default_retry, timeout=self.C.TIMEOUT)

    def get_file_load(self, file_path: str) -> LoadJob:
        try:
            return BigQueryLoadJob(
                SqlClientBase.get_file_name_from_file_path(file_path),
                self._retrieve_load_job(file_path),
                self.C
            )
        except api_core_exceptions.NotFound:
            raise LoadJobNotExistsException(file_path)
        except (api_core_exceptions.BadRequest, api_core_exceptions.NotFound):
            raise LoadJobServerTerminalException(file_path)

    def start_file_load(self, table_name: str, file_path: str) -> LoadJob:
        # verify that table exists in the schema
        self._get_table_by_name(table_name, file_path)
        try:
            return BigQueryLoadJob(
                SqlClientBase.get_file_name_from_file_path(file_path),
                self._create_load_job(table_name, file_path),
                self.C
            )
        except api_core_exceptions.NotFound:
            # google.api_core.exceptions.BadRequest - will not be processed ie bad job name
            raise LoadUnknownTableException(table_name, file_path)
        except (api_core_exceptions.BadRequest, api_core_exceptions.NotFound):
            # google.api_core.exceptions.NotFound: 404 - table not found
            raise LoadJobServerTerminalException(file_path)
        except api_core_exceptions.Conflict:
            # google.api_core.exceptions.Conflict: 409 PUT - already exists
            return self.get_file_load(file_path)

    def update_storage_schema(self) -> None:
        storage_version = self._get_schema_version_from_storage()
        if storage_version < self.schema.schema_version:
            for sql in self._build_schema_update_sql():
                self._execute_sql(sql)
            self._update_schema_version(self.schema.schema_version)

    def _open_connection(self) -> None:
        credentials = service_account.Credentials.from_service_account_info(self.C.to_service_credentials())
        self._client = bigquery.Client(self.C.PROJECT_ID, credentials=credentials)

    def _close_connection(self) -> None:
        if self._client:
            self._client.close()
            self._client = None

    def _get_schema_version_from_storage(self) -> int:
        try:
            return super()._get_schema_version_from_storage()
        except api_core_exceptions.NotFound:
            # there's no table so there's no schema
            return 0

    def _build_schema_update_sql(self) -> List[str]:
        sql_updates = []
        for table_name in self.schema.schema_tables:
            exists, storage_table = self._get_storage_table(table_name)
            sql = self._get_table_update_sql(table_name, storage_table, exists)
            if sql:
                sql_updates.append(sql)
        return sql_updates

    def _get_table_update_sql(self, table_name: str, storage_table: Table, exists: bool) -> str:
        new_columns = self._create_table_update(table_name, storage_table)
        if len(new_columns) == 0:
            # no changes
            return None
        # build sql
        canonical_name = self._to_canonical_table_name(table_name)
        if not exists:
            # build CREATE
            sql = f"CREATE TABLE {canonical_name} (\n"
            sql += ",\n".join([self._get_column_def_sql(c) for c in new_columns])
            sql += ")"
        else:
            # build ALTER
            sql = f"ALTER TABLE {canonical_name}\n"
            sql += ",\n".join(["ADD COLUMN " + self._get_column_def_sql(c) for c in new_columns])
        # scan columns to get hints
        cluster_list = [escape_bigquery_identifier(c["name"]) for c in new_columns if c.get("cluster", False)]
        partition_list = [escape_bigquery_identifier(c["name"]) for c in new_columns if c.get("partition", False)]
        # partition by must be added first
        if len(partition_list) > 0:
            if exists:
                raise LoadClientSchemaWillNotUpdate(canonical_name, partition_list, "Partition requested after table was created")
            elif len(partition_list) > 1:
                raise LoadClientSchemaWillNotUpdate(canonical_name, partition_list, "Partition requested for more than one column")
            else:
                sql += f"\nPARTITION BY DATE({partition_list[0]})"
        if len(cluster_list) > 0:
            if exists:
                raise LoadClientSchemaWillNotUpdate(canonical_name, cluster_list, "Clustering requested after table was created")
            else:
                sql += "\nCLUSTER BY " + ",".join(cluster_list)

        return sql

    def _get_column_def_sql(self, c: Column) -> str:
        name = escape_bigquery_identifier(c["name"])
        return f"{name} {self._sc_t_to_bq_t(c['data_type'])} {self._gen_not_null(c['nullable'])}"

    def _get_storage_table(self, table_name: str) -> Tuple[bool, Table]:
        schema_table: Table = {}
        try:
            table = self._client.get_table(self._to_canonical_table_name(table_name), retry=self.default_retry, timeout=self.C.TIMEOUT)
            partition_field = table.time_partitioning.field if table.time_partitioning else None
            for c in table.schema:
                schema_c: Column = {
                    "name": c.name,
                    "nullable": c.is_nullable,
                    "data_type": self._bq_t_to_sc_t(c.field_type, c.precision, c.scale),
                    "unique": False,
                    "sort": False,
                    "primary_key": False,
                    "foreign_key": False,
                    "cluster": c.name in (table.clustering_fields or []),
                    "partition": c.name == partition_field
                }
                schema_table[c.name] = schema_c
            return True, schema_table
        except gcp_exceptions.NotFound:
            return False, schema_table

    def _execute_sql(self, query: AnyStr) -> Any:
        logger.debug(f"Will execute query {query}")  # type: ignore
        return self._client.query(query, job_retry=self.default_retry, timeout=self.C.TIMEOUT).result()

    def _to_canonical_schema_name(self) -> str:
        return f"{self.C.PROJECT_ID}.{self.C.DATASET}_{self.schema.schema_name}"

    def _create_load_job(self, table_name: str, file_path: str) -> bigquery.LoadJob:
        job_id = BigQueryClient._get_job_id_from_file_path(file_path)
        job_config = bigquery.LoadJobConfig(
            autodetect=False,
            write_disposition=bigquery.WriteDisposition.WRITE_APPEND,
            create_disposition=bigquery.CreateDisposition.CREATE_NEVER,
            source_format=bigquery.SourceFormat.NEWLINE_DELIMITED_JSON,
            ignore_unknown_values=False,
            max_bad_records=0,

            )
        with open(file_path, "rb") as f:
            return self._client.load_table_from_file(f,
                                                     self._to_canonical_table_name(table_name),
                                                     job_id=job_id,
                                                     job_config=job_config,
                                                     timeout=self.C.TIMEOUT
                                                    )

    def _retrieve_load_job(self, file_path: str) -> bigquery.LoadJob:
        job_id = BigQueryClient._get_job_id_from_file_path(file_path)
        return self._client.get_job(job_id)

    @staticmethod
    def _get_job_id_from_file_path(file_path: str) -> str:
        return Path(file_path).name.replace(".", "_")

    @staticmethod
    def _gen_not_null(v: bool) -> str:
        return "NOT NULL" if not v else ""

    @staticmethod
    def _sc_t_to_bq_t(sc_t: DataType) -> str:
        return SCT_TO_BQT[sc_t]

    @staticmethod
    def _bq_t_to_sc_t(bq_t: str, precision: Optional[int], scale: Optional[int]) -> DataType:
        if bq_t == "BIGNUMERIC":
            if precision is None:  # biggest numeric possible
                return "wei"
        return BQT_TO_SCT.get(bq_t, "text")


def make_client(schema: Schema, C: Type[GcpClientConfiguration]) -> BigQueryClient:
    return BigQueryClient(schema, C)


def supported_writer(C: Type[GcpClientConfiguration]) -> TWriterType:
    return "jsonl"

# cred = service_account.Credentials.from_service_account_info(_credentials)
# project_id = cred.get('project_id')
# client = bigquery.Client(project_id, credentials=cred)
# print(client.get_dataset("carbon_bot_extract_7"))
# exit(0)
# from dlt.common.configuration import SchemaStoreConfiguration
# from dlt.common.logger import  init_logging_from_config

# init_logging_from_config(CLIENT_CONFIG)

# schema = Schema(SchemaStoreConfiguration.TRACKER_SCHEMA_FILE_PATH)
# schema.load_schema()
# import pprint
# # pprint.pprint(schema.as_yaml())
# with make_client(schema) as client:
#     client.initialize_storage()
#     # job = client._create_load_job("tracker", "_storage/loaded/1630949263.574516/completed_jobs/tracker.1c31ff1b-c250-4690-8973-14f0ee9ae355.jsonl")
#     # unk table
#     # job = client._create_load_job("trackerZ", "_storage/loaded/1630949263.574516/completed_jobs/tracker.4876f905-aefe-4262-a440-d29ed2643c3a.jsonl")
#     # job = client._create_load_job("tracker", "_storage/loaded/1630949263.574516/completed_jobs/event_bot.c9105079-2d1d-4ad3-8613-a5dff790889d.jsonl")
#     # failed
#     # job = client._retrieve_load_job("_storage/loaded/1630949263.574516/completed_jobs/event_bot.c9105079-2d1d-4ad3-8613-a5dff790889d.jsonl")
#     # OK
#     job = client._retrieve_load_job("_storage/loaded/1630949263.574516/completed_jobs/tracker.1c31ff1b-c250-4690-8973-14f0ee9ae355.jsonl")
#     while True:
#         try:
#             # this does not throw
#             done = job.done()
#             print(f"DONE: {job.done(reload=False)}")
#         except Exception as e:
#             logger.exception("DONE")
#             done = True
#         if done:
#             break;
#         # done is not self running

#         # print(job.running())
#         sleep(1)
#     try:
#         print(f"status: {job.state}")
#         print(f"error: {job.error_result}")
#         print(f"errors: {job.errors}")
#         print(f"line count: {job.output_rows}")
#         print(job.exception())
#     except:
#         logger.exception("EXCEPTION")
#     try:
#         print(job.result())
#     except:
#         logger.exception("RESULT")

    # non existing table
    # wrong data - unknown column




import jsonlines
from typing import Any, Iterable, Literal, Sequence, IO

from dlt.common import json
from dlt.common.typing import StrAny

TWriterType = Literal["jsonl", "insert_values"]

def write_jsonl(f: IO[Any], rows: Sequence[Any]) -> None:
    # use jsonl to write load files https://jsonlines.org/
    with jsonlines.Writer(f, dumps=json.dumps) as w:
        w.write_all(rows)


def write_insert_values(f: IO[Any], rows: Sequence[StrAny], headers: Iterable[str]) -> None:
    # dict lookup is always faster
    headers_lookup = {v: i for i, v in enumerate(headers)}
    # do not write INSERT INTO command, this must be added together with table name by the loader
    f.write("INSERT INTO {}(")
    f.write(",".join(map(escape_redshift_identifier, headers)))
    f.write(")\nVALUES\n")

    def stringify(v: Any) -> str:
        if isinstance(v, bytes):
           return f"from_hex('{v.hex()}')"
        else:
            return str(v)

    def write_row(row: StrAny) -> None:
        output = ["NULL" for _ in range(len(headers_lookup))]
        for n,v  in row.items():
            output[headers_lookup[n]] = escape_redshift_literal(v) if isinstance(v, str) else stringify(v)
        f.write("(")
        f.write(",".join(output))
        f.write(")")

    for row in rows[:-1]:
        write_row(row)
        f.write(",\n")

    write_row(rows[-1])
    f.write(";")


def escape_redshift_literal(v: str) -> str:
    # https://www.postgresql.org/docs/9.3/sql-syntax-lexical.html
    # looks like this is the only thing we need to escape for Postgres > 9.1
    # redshift keeps \ as escape character which is pre 9 behavior
    return "'" + v.replace("'", "''").replace("\\", "\\\\") + "'"


def escape_redshift_identifier(v: str) -> str:
    return '"' + v.replace('"', '""').replace("\\", "\\\\") + '"'


def escape_bigquery_identifier(v: str) -> str:
    # https://cloud.google.com/bigquery/docs/reference/standard-sql/lexical
    return "`" + v.replace("\\", "\\\\").replace("`","\\`") + "`"



import semver
from dlt.common.exceptions import DltException

class StorageException(DltException):
    def __init__(self, msg: str) -> None:
        super().__init__(msg)


class NoMigrationPathException(StorageException):
    def __init__(self, storage_path: str, initial_version: semver.VersionInfo, migrated_version: semver.VersionInfo, target_version: semver.VersionInfo) -> None:
        self.storage_path = storage_path
        self.initial_version = initial_version
        self.migrated_version = migrated_version
        self.target_version = target_version
        super().__init__(f"Could not find migration path for {storage_path} from v {initial_version} to {target_version}, stopped at {migrated_version}")


class WrongStorageVersionException(StorageException):
    def __init__(self, storage_path: str, initial_version: semver.VersionInfo, target_version: semver.VersionInfo) -> None:
        self.storage_path = storage_path
        self.initial_version = initial_version
        self.target_version = target_version
        super().__init__(f"Expected storage {storage_path} with v {target_version} but found {initial_version}")



from typing import Sequence
from dlt.common.exceptions import DltException
from dlt.common.telemetry import TRunMetrics
from dlt.pipeline.typing import TPipelineStage


class PipelineException(DltException):
    pass


class MissingDependencyException(PipelineException):
    def __init__(self, caller: str, dependencies: Sequence[str], appendix: str = "") -> None:
        self.caller = caller
        self.dependencies = dependencies
        super().__init__(self._get_msg(appendix))

    def _get_msg(self, appendix: str) -> str:
        msg = f"""
You must install additional dependencies to run {self.caller}. If you use pip you may do the following:

{self._to_pip_install()}
"""
        if appendix:
            msg = msg + "\n" + appendix
        return msg

    def _to_pip_install(self) -> str:
        return "\n".join([f"pip install {d}" for d in self.dependencies])


class NoPipelineException(PipelineException):
    def __init__(self) -> None:
        super().__init__("Please create or restore pipeline before using this function")


class InvalidPipelineContextException(PipelineException):
    def __init__(self) -> None:
        super().__init__("There may be just one active pipeline in single python process. You may have switch between pipelines by restoring pipeline just before using load method")


class CannotRestorePipelineException(PipelineException):
    def __init__(self, reason: str) -> None:
        super().__init__(reason)


class PipelineStepFailed(PipelineException):
    def __init__(self, stage: TPipelineStage, exception: BaseException, run_metrics: TRunMetrics) -> None:
        self.stage = stage
        self.exception = exception
        self.run_metrics = run_metrics
        super().__init__(f"Pipeline execution failed at stage {stage} with exception:\n\n{type(exception)}\n{exception}")



from dlt.common.schema import Schema

from dlt.pipeline import Pipeline


def discover_schema() -> Schema:
    # all Ethereum compatible blockchains have the same schema so we just provide a nice yaml file
    return Pipeline.load_schema_from_file("examples/schemas/ethereum_schema.yml")



import os

from dlt.pipeline.pipeline import Pipeline
from dlt.pipeline.typing import PostgresPipelineCredentials

from examples.sources.sql_query import get_source

credentials = PostgresPipelineCredentials("redshift", "chat_analytics_rasa", "gamma_guild_7", "loader", "chat-analytics.czwteevq7bpe.eu-central-1.redshift.amazonaws.com")

# create sql alchemy connector
conn_str = f"redshift+redshift_connector://{credentials.PG_USER}:{os.environ['PG_PASSWORD']}@{credentials.PG_HOST}:{credentials.PG_PORT}/{credentials.PG_DATABASE_NAME}"

# items = get_source("select *  from test_fixture_carbon_bot_session_cases_views.users limit 1", conn_str)


# we preserve method signature from pandas
items = get_source("select *  from mainnet_2_ethereum.blocks__transactions limit 10", conn_str, coerce_float=False)

# for i in items:
#     for k, v in i.items():
#         print(f"{k}:{v} ({type(v)}:{Schema._py_type_to_sc_type(type(v))})")

# unpack and display schema
p = Pipeline("mydb")
p.create_pipeline(credentials)
p.extract(items, table_name="blocks__transactions")
p.unpack()
print(p.get_default_schema().as_yaml(remove_default_hints=True))




from dataclasses import dataclass
from typing import Callable, Dict, Iterable, Iterator, List, Literal, Sequence, Tuple, TypeVar, Union, Generic
from dlt.common import json

from dlt.common.typing import DictStrAny, StrAny
from dlt.common.configuration.utils import TConfigSecret

TLoaderType = Literal["gcp", "redshift"]
TPipelineStage = Literal["extract", "unpack", "load"]

# extractor generator yields functions that returns list of items of the type (table) when called
# this allows generator to implement retry logic
# TExtractorItem = Callable[[], Iterator[StrAny]]
# # extractor generator yields tuples: (type of the item (table name), function defined above)
# TExtractorItemWithTable = Tuple[str, TExtractorItem]
# TExtractorGenerator = Callable[[DictStrAny], Iterator[TExtractorItemWithTable]]


@dataclass
class PipelineCredentials:
    CLIENT_TYPE: TLoaderType


@dataclass
class GCPPipelineCredentials(PipelineCredentials):
    PROJECT_ID: str
    DATASET: str
    BQ_CRED_CLIENT_EMAIL: str
    BQ_CRED_PRIVATE_KEY: TConfigSecret = None
    TIMEOUT: float = 30.0

    @classmethod
    def from_services_dict(cls, services: StrAny, dataset_prefix: str) -> "GCPPipelineCredentials":
        assert dataset_prefix is not None

        return cls("gcp", services["project_id"], dataset_prefix, services["client_email"], services["private_key"])

    @classmethod
    def from_services_file(cls, services_path: str, dataset_prefix: str) -> "GCPPipelineCredentials":
        with open(services_path, "r") as f:
            services = json.load(f)
        return GCPPipelineCredentials.from_services_dict(services, dataset_prefix)


@dataclass
class PostgresPipelineCredentials(PipelineCredentials):
    PG_DATABASE_NAME: str
    PG_SCHEMA_PREFIX: str
    PG_USER: str
    PG_HOST: str
    PG_PASSWORD: TConfigSecret = None
    PG_PORT: int = 5439
    PG_CONNECTION_TIMEOUT: int = 15



from typing import Any, Callable, Type, List, Dict, Optional, Sequence, Tuple
from multiprocessing.pool import Pool as ProcessPool
from itertools import chain
from prometheus_client import Counter, CollectorRegistry, REGISTRY, Gauge
from prometheus_client.metrics import MetricWrapperBase

from dlt.common import pendulum, signals, json, logger
from dlt.common.json import custom_pua_decode
from dlt.common.runners import TRunArgs, TRunMetrics, create_default_args, pool_runner, initialize_runner
from dlt.common.storages.unpacker_storage import UnpackerStorage
from dlt.common.telemetry import get_logging_extras
from dlt.common.utils import uniq_id
from dlt.common.typing import TEvent
from dlt.common.logger import process_internal_exception
from dlt.common.exceptions import PoolException
from dlt.common.storages import SchemaStorage
from dlt.common.schema import CannotCoerceColumnException, SchemaUpdate, Schema
from dlt.common.parser import PATH_SEPARATOR
from dlt.common.storages.loader_storage import LoaderStorage

from dlt.common.parser import extract, TExtractFunc
from dlt.unpacker.configuration import configuration, UnpackerConfiguration

extract_func: TExtractFunc = extract
CONFIG: Type[UnpackerConfiguration] = None
unpack_storage: UnpackerStorage = None
load_storage: LoaderStorage = None
schema_storage: SchemaStorage = None
load_schema_storage: SchemaStorage = None
event_counter: Counter = None
event_gauge: Gauge = None
schema_version_gauge: Gauge = None
load_package_counter: Counter = None


def create_gauges(registry: CollectorRegistry) -> Tuple[MetricWrapperBase, MetricWrapperBase, MetricWrapperBase, MetricWrapperBase]:
    return (
        Counter("unpacker_event_count", "Events processed in unpacker", ["schema"], registry=registry),
        Gauge("unpacker_last_events", "Number of events processed in last run", ["schema"], registry=registry),
        Gauge("unpacker_schema_version", "Current schema version", ["schema"], registry=registry),
        Gauge("unpacker_load_packages_created_count", "Count of load package created", ["schema"], registry=registry)
    )


def create_folders() -> Tuple[UnpackerStorage, LoaderStorage, SchemaStorage, SchemaStorage]:
    unpack_storage = UnpackerStorage(True, CONFIG)
    schema_storage = SchemaStorage(CONFIG.SCHEMA_VOLUME_PATH, makedirs=True)
    load_schema_storage = SchemaStorage(CONFIG.LOADING_VOLUME_PATH, makedirs=False)
    load_storage = LoaderStorage(True, CONFIG, CONFIG.WRITER_TYPE)

    unpack_storage.initialize_storage()
    load_storage.initialize_storage()

    return unpack_storage, load_storage, schema_storage, load_schema_storage


def install_schemas(default_schemas_path: str, schema_names: List[str]) -> None:
    # copy default schemas if not present
    default_schemas = SchemaStorage(default_schemas_path)
    logger.info(f"Checking default schemas in {schema_storage.storage.storage_path}")
    for name in schema_names:
        if not schema_storage.has_store_schema(name):
            logger.info(f"Schema, {name} not present in {schema_storage.storage.storage_path}, installing...")
            schema = default_schemas.load_store_schema(name)
            schema_storage.save_store_schema(schema)


def load_or_create_schema(schema_name: str) -> Schema:
    try:
        schema = schema_storage.load_store_schema(schema_name)
        logger.info(f"Loaded schema with name {schema_name} with version {schema.schema_version}")
    except FileNotFoundError:
        schema = Schema(schema_name)
        logger.info(f"Created new schema with name {schema_name}")
    return schema


# this is a worker process
def w_unpack_files(schema_name: str, load_id: str, events_files: Sequence[str]) -> SchemaUpdate:
    unpacked_data: Dict[str, List[Any]] = {}

    schema_update: SchemaUpdate = {}
    schema = load_or_create_schema(schema_name)
    file_id = uniq_id()

    # process all event files and store rows in memory
    for events_file in events_files:
        try:
            logger.debug(f"Processing events file {events_file} in load_id {load_id} with file_id {file_id}")
            with unpack_storage.storage.open(events_file) as f:
                events: Sequence[TEvent] = json.load(f)
            for event in events:
                for table_name, row in extract_func(schema, event, load_id, CONFIG.ADD_EVENT_JSON):
                    # filter row, may eliminate some or all fields
                    row = schema.filter_row(table_name, row, PATH_SEPARATOR)
                    # do not process empty rows
                    if row:
                        # decode pua types
                        for k, v in row.items():
                            row[k] = custom_pua_decode(v)  # type: ignore
                        # check if schema can be updated
                        row, table_update = schema.coerce_row(table_name, row)
                        if len(table_update) > 0:
                            # update schema and save the change
                            schema.update_schema(table_name, table_update)
                            table_updates = schema_update.setdefault(table_name, [])
                            table_updates.extend(table_update)
                        # store row
                        rows = unpacked_data.setdefault(table_name, [])
                        rows.append(row)
        except Exception:
            process_internal_exception(f"Exception when processing file {events_file}")
            raise PoolException("unpack_files", events_file)

    # save rows and return schema changes to be gathered in parent process
    for table_name, rows in unpacked_data.items():
        # save into new jobs to processed as load
        table = schema.get_table(table_name)
        load_storage.write_temp_loading_file(load_id, table_name, table, file_id, rows)

    return schema_update


TMapFuncRV = Tuple[List[SchemaUpdate], List[Sequence[str]]]
TMapFuncType = Callable[[ProcessPool, str, str, Sequence[str]], TMapFuncRV]

def map_parallel(pool: ProcessPool, schema_name: str, load_id: str, files: Sequence[str]) -> TMapFuncRV:
    # we chunk files in a way to not exceed MAX_EVENTS_IN_CHUNK and split them equally
    # between processors
    configured_processes = pool._processes  # type: ignore
    chunk_files = UnpackerStorage.chunk_by_events(files, CONFIG.MAX_EVENTS_IN_CHUNK, configured_processes)
    logger.info(f"Obtained {len(chunk_files)} processing chunks")
    param_chunk = [(schema_name, load_id, files) for files in chunk_files]
    return pool.starmap(w_unpack_files, param_chunk), chunk_files


def map_single(_: ProcessPool, schema_name: str, load_id: str, files: Sequence[str]) -> TMapFuncRV:
    chunk_files = UnpackerStorage.chunk_by_events(files, CONFIG.MAX_EVENTS_IN_CHUNK, 1)
    # get in one chunk
    assert len(chunk_files) == 1
    logger.info(f"Obtained {len(chunk_files)} processing chunks")
    return [w_unpack_files(schema_name, load_id, chunk_files[0])], chunk_files


def update_schema(schema_name: str, schema_updates: List[SchemaUpdate]) -> Schema:
    schema = load_or_create_schema(schema_name)
    # gather schema from all manifests, validate consistency and combine
    for schema_update in schema_updates:
        for table_name, table_updates in schema_update.items():
            logger.debug(f"Updating schema for table {table_name} with {len(table_updates)} deltas")
            schema.update_schema(table_name, table_updates)
    return schema


def spool_files(pool: ProcessPool, schema_name: str, load_id: str, map_f: TMapFuncType, files: Sequence[str]) -> None:
    # process files in parallel or in single thread, depending on map_f
    schema_updates, chunk_files = map_f(pool, schema_name, load_id, files)

    schema = update_schema(schema_name, schema_updates)
    schema_version_gauge.labels(schema_name).set(schema._version)
    logger.metrics("Unpacker metrics", extra=get_logging_extras([schema_version_gauge.labels(schema_name)]))
    logger.info(f"Saving schema {schema_name} with version {schema._version}, writing manifest files")
    # schema is updated, save it to schema volume
    schema_storage.save_store_schema(schema)
    # save schema and schema updates to temp load folder
    load_schema_storage.save_folder_schema(schema, load_id)
    load_storage.save_schema_updates(load_id, schema_updates)
    # files must be renamed and deleted together so do not attempt that when process is about to be terminated
    signals.raise_if_signalled()
    logger.info(f"Committing storage, do not kill this process")
    # rename temp folder to processing
    load_storage.commit_temp_load_folder(load_id)
    # delete event files and count events to provide metrics
    total_events = 0
    for event_file in chain.from_iterable(chunk_files):  # flatten chunks
        unpack_storage.storage.delete(event_file)
        total_events += UnpackerStorage.get_events_count(event_file)
    # log and update metrics
    logger.info(f"Chunk {load_id} processed")
    load_package_counter.labels(schema_name).inc()
    event_counter.labels(schema_name).inc(total_events)
    event_gauge.labels(schema_name).set(total_events)
    logger.metrics("Unpacker metrics", extra=get_logging_extras(
        [load_package_counter.labels(schema_name), event_counter.labels(schema_name), event_gauge.labels(schema_name)]))


def spool_schema_files(pool: ProcessPool, schema_name: str, files: Sequence[str]) -> str:
    # unpacked files will go here before being atomically renamed
    load_id = str(pendulum.now().timestamp())
    load_storage.create_temp_load_folder(load_id)
    logger.info(f"Created temp load folder {load_id} on loading volume")

    try:
        # process parallel
        spool_files(pool, schema_name, load_id, map_parallel, files)
    except CannotCoerceColumnException as exc:
        # schema conflicts resulting from parallel executing
        logger.warning(f"Parallel schema update conflict, switching to single thread ({str(exc)}")
        # start from scratch
        load_storage.create_temp_load_folder(load_id)
        spool_files(pool, schema_name, load_id, map_single, files)

    return load_id


def unpack(pool: ProcessPool) -> TRunMetrics:
    logger.info(f"Running file unpacking")
    # list files and group by schema name, list must be sorted for group by to actually work
    files = unpack_storage.list_files_to_unpack_sorted()
    logger.info(f"Found {len(files)} files, will process in chunks of {CONFIG.MAX_EVENTS_IN_CHUNK} of events")
    if len(files) == 0:
        return TRunMetrics(True, False, 0)
    # group files by schema
    for schema_name, files_in_schema in unpack_storage.get_grouped_iterator(files):
        logger.info(f"Found files in schema {schema_name}")
        spool_schema_files(pool, schema_name, list(files_in_schema))
    # return info on still pending files (if extractor saved something in the meantime)
    return TRunMetrics(False, False, len(unpack_storage.list_files_to_unpack_sorted()))


def configure(C: Type[UnpackerConfiguration], collector: CollectorRegistry, extract_f: TExtractFunc, default_schemas_path: str = None, schema_names: List[str] = None) -> None:
    global CONFIG
    global unpack_storage, load_storage, schema_storage, load_schema_storage
    global event_counter, event_gauge, schema_version_gauge, load_package_counter
    global extract_func

    CONFIG = C
    # set extracting parser function
    extract_func = extract_f
    # setup singletons
    unpack_storage, load_storage, schema_storage, load_schema_storage = create_folders()
    try:
        event_counter, event_gauge, schema_version_gauge, load_package_counter = create_gauges(collector)
    except ValueError as v:
        # ignore re-creation of gauges
        if "Duplicated timeseries" not in str(v):
            raise
    if default_schemas_path and schema_names:
        install_schemas(default_schemas_path, schema_names)


def main(args: TRunArgs, extract_f: TExtractFunc, default_schemas_path: str = None, schema_names: List[str] = None) -> int:
    # initialize runner
    C = configuration()
    initialize_runner(C, args)
    # create objects and gauges
    try:
        configure(C, REGISTRY, extract_f, default_schemas_path, schema_names)
    except Exception:
        process_internal_exception("init module")
        return -1
    # unpack
    return pool_runner(C, unpack)


def run_main(args: TRunArgs) -> None:
    exit(main(args, extract))



from dlt.pipeline import Pipeline, PostgresPipelineCredentials

credentials = PostgresPipelineCredentials("redshift", "chat_analytics_rasa", "mainnet_3", "loader", "3.73.90.3")
# credentials = Pipeline.load_gcp_credentials("_secrets/project1234_service.json", "mainnet_4")
pipeline = Pipeline("ethereum")
pipeline.create_pipeline(credentials)
schema = Pipeline.load_schema_from_file("examples/schemas/ethereum_schema.yml")
# set the loaded schema for the whole pipeline
pipeline.set_default_schema(schema)
# will sync schema with the target
pipeline.sync_schema()



import logging
import json_logging
import traceback
import sentry_sdk
from sentry_sdk.transport import HttpTransport
from logging import LogRecord, Logger
from typing import Any, Callable, Dict, Type

from dlt.common.json import json
from dlt.common.typing import DictStrAny, DictStrStr, StrStr
from dlt.common.configuration import BasicConfiguration
from dlt.common.utils import filter_env_vars
from dlt._version import common_version as __version__

DLT_LOGGER_NAME = "sv-dlt"
LOGGER: Logger = None

def _add_logging_level(level_name: str, level: int, method_name:str = None) -> None:
    """
    Comprehensively adds a new logging level to the `logging` module and the
    currently configured logging class.

    `levelName` becomes an attribute of the `logging` module with the value
    `levelNum`. `methodName` becomes a convenience method for both `logging`
    itself and the class returned by `logging.getLoggerClass()` (usually just
    `logging.Logger`). If `methodName` is not specified, `levelName.lower()` is
    used.

    To avoid accidental clobberings of existing attributes, this method will
    raise an `AttributeError` if the level name is already an attribute of the
    `logging` module or if the method name is already present

    """
    if not method_name:
        method_name = level_name.lower()

    if hasattr(logging, level_name):
       raise AttributeError('{} already defined in logging module'.format(level_name))
    if hasattr(logging, method_name):
       raise AttributeError('{} already defined in logging module'.format(method_name))
    if hasattr(logging.getLoggerClass(), method_name):
       raise AttributeError('{} already defined in logger class'.format(method_name))

    # This method was inspired by the answers to Stack Overflow post
    # http://stackoverflow.com/q/2183233/2988730, especially
    # http://stackoverflow.com/a/13638084/2988730
    def logForLevel(self: logging.Logger, message: str, *args: Any, **kwargs: Any) -> None:
        if self.isEnabledFor(level):
            self._log(level, message, args, **kwargs)
    def logToRoot(message: str, *args: Any, **kwargs: Any) -> None:
        logging.root._log(level, message, args, **kwargs)

    logging.addLevelName(level, level_name)
    setattr(logging, level_name, level)
    setattr(logging.getLoggerClass(), method_name, logForLevel)
    setattr(logging, method_name, logToRoot)


class _MetricsFormatter(logging.Formatter):
    def format(self, record: LogRecord) -> str:
        s = super(_MetricsFormatter, self).format(record)
        if record.exc_text:
            s = s + '|'
        # dump metrics dictionary nicely
        if "metrics" in record.__dict__:
            s = s + ": " + json.dumps(record.__dict__["metrics"])
        return s


class _CustomJsonFormatter(json_logging.JSONLogFormatter):

    version: StrStr = None

    def _format_log_object(self, record: LogRecord, request_util: Any) -> Any:
        json_log_object = super(_CustomJsonFormatter, self)._format_log_object(record, request_util)
        if self.version:
            json_log_object.update({"version": self.version})
        return json_log_object


def _init_logging(logger_name: str, level: str, format: str, component: str, version: StrStr) -> Logger:
    if logger_name == "root":
        logging.basicConfig(level=level)
        handler = logging.getLogger().handlers[0]
        # handler.setFormatter(_MetricsFormatter(fmt=format, style='{'))
        logger = logging.getLogger()
    else:
        logger = logging.getLogger(DLT_LOGGER_NAME)
        logger.propagate = False
        logger.setLevel(level)
        # get or create logging handler
        handler = next(iter(logger.handlers), logging.StreamHandler())
        logger.addHandler(handler)

    # set right formatter
    if is_json_logging(format):
        json_logging.COMPONENT_NAME = component
        json_logging.JSON_SERIALIZER = json.dumps
        json_logging.RECORD_ATTR_SKIP_LIST.remove("process")
        # set version as class variable as we cannot pass custom constructor parameters
        _CustomJsonFormatter.version = version
        # the only thing method above effectively does is to replace the formatter
        json_logging.init_non_web(enable_json=True, custom_formatter=_CustomJsonFormatter)
        if logger_name == "root":
            json_logging.config_root_logger()
    else:
        handler.setFormatter(_MetricsFormatter(fmt=format, style='{'))

    return logger


def __getattr__(name: str) -> Callable[..., Any]:
    # a catch all function for a module that forwards calls to unknown methods to LOGGER
    def wrapper(msg: str, *args: Any, **kwargs: Any) -> None:
        if LOGGER:
            getattr(LOGGER, name)(msg, *args, **kwargs, stacklevel=2)
    return wrapper


def _extract_version_info(config: Type[BasicConfiguration]) -> StrStr:
    version_info = {"version": __version__, "component_name": config.NAME}
    version = getattr(config, "_VERSION", None)
    if version:
        version_info["component_version"] = version
    # extract envs with build info
    version_info.update(filter_env_vars(["COMMIT_SHA", "IMAGE_VERSION"]))
    return version_info


def _extract_pod_info() -> StrStr:
    return filter_env_vars(["KUBE_NODE_NAME", "KUBE_POD_NAME", "KUBE_POD_NAMESPACE"])

class _SentryHttpTransport(HttpTransport):

    timeout: int = 0

    def _get_pool_options(self, *a: Any, **kw: Any) -> DictStrAny:
        rv = HttpTransport._get_pool_options(self, *a, **kw)
        rv['timeout'] = self.timeout
        return rv


def _init_sentry(config: Type[BasicConfiguration], version: StrStr) -> None:
    if config.SENTRY_DSN:
        sys_ver = version["version"]
        release = sys_ver + "_" + version.get("commit_sha", "")
        _SentryHttpTransport.timeout = config.REQUEST_TIMEOUT[0]
        # TODO: setup automatic sending of log messages by log level (ie. we send a lot dbt trash logs)
        # https://docs.sentry.io/platforms/python/guides/logging/
        sentry_sdk.init(config.SENTRY_DSN, release=release, transport=_SentryHttpTransport)
        # add version tags
        for k, v in version.items():
            sentry_sdk.set_tag(k, v)
        # add kubernetes tags
        pod_tags = _extract_pod_info()
        for k, v in pod_tags.items():
            sentry_sdk.set_tag(k, v)


def init_telemetry(config: Type[BasicConfiguration]) -> None:
    if config.PROMETHEUS_PORT:
        from prometheus_client import start_http_server, Info

        logging.info(f"Starting prometheus server port {config.PROMETHEUS_PORT}")
        start_http_server(config.PROMETHEUS_PORT)
        # collect info
        Info("runs_component_name", "Name of the executing component").info(_extract_version_info(config))


def init_logging_from_config(C: Type[BasicConfiguration]) -> None:
    global LOGGER

    # add HEALTH and METRICS log levels
    if not hasattr(logging, "health"):
        _add_logging_level("HEALTH", logging.WARNING - 1, "health")
        _add_logging_level("METRICS", logging.WARNING - 2, "metrics")

    version = _extract_version_info(C)
    LOGGER = _init_logging(
        DLT_LOGGER_NAME,
        C.LOG_LEVEL,
        C.LOG_FORMAT,
        C.NAME,
        version)
    _init_sentry(C, version)


def is_json_logging(log_format: str) -> bool:
    return log_format == "JSON"


def process_internal_exception(msg: str, exc_info: Any = True) -> None:
    # Passing default True value will cause implementation to use data provided by sys.exc_info
    if LOGGER:
        LOGGER.error(msg, exc_info=exc_info, stacklevel=2)
    report_exception()


def report_exception() -> None:
    if sentry_sdk.Hub.current:
        sentry_sdk.capture_exception()


def pretty_format_exception() -> str:
    return traceback.format_exc()



from abc import ABC, abstractmethod
from types import TracebackType
from typing import Any, Literal, Sequence, Type, TypeVar, AnyStr
from pathlib import Path

from dlt.common import pendulum, logger
from dlt.common.schema import Column, Schema, Table
# from dlt.common.file_storage import FileStorage

from dlt.loaders.local_types import LoadJobStatus
from dlt.loaders.exceptions import LoadClientSchemaVersionCorrupted, LoadUnknownTableException

# typing for context manager
TClient = TypeVar("TClient", bound="ClientBase")


class LoadJob:
    def __init__(self, file_name: str) -> None:
        """
        File name is also a job id (or job id is deterministically derived) so it must be globally unique
        """
        self._file_name = file_name

    @abstractmethod
    def status(self) -> LoadJobStatus:
        pass

    @abstractmethod
    def file_name(self) -> str:
        pass

    @abstractmethod
    def exception(self) -> str:
        pass


class LoadEmptyJob(LoadJob):
    def __init__(self, file_name: str, status: LoadJobStatus, exception: str = None) -> None:
        self._status = status
        self._exception = exception
        super().__init__(file_name)

    def status(self) -> LoadJobStatus:
        return self._status

    def file_name(self) -> str:
        return self._file_name

    def exception(self) -> str:
        return self._exception


class ClientBase(ABC):
    def __init__(self, schema: Schema) -> None:
        self.schema = schema

    @abstractmethod
    def initialize_storage(self) -> None:
        pass

    @abstractmethod
    def update_storage_schema(self) -> None:
        pass

    @abstractmethod
    def start_file_load(self, table_name: str, file_path: str) -> LoadJob:
        pass

    @abstractmethod
    def get_file_load(self, file_path: str) -> LoadJob:
        pass

    @abstractmethod
    def complete_load(self, load_id: str) -> None:
        pass

    @abstractmethod
    def _open_connection(self) -> None:
        pass

    @abstractmethod
    def _close_connection(self) -> None:
        pass


    def __enter__(self: TClient) -> TClient:
        self._open_connection()
        return self

    def __exit__(self, exc_type: Type[BaseException], exc_val: BaseException, exc_tb: TracebackType) -> None:
        self._close_connection()

    def _get_table_by_name(self, table_name: str, file_name: str) -> Table:
        try:
            return self.schema.get_table(table_name)
        except KeyError:
            raise LoadUnknownTableException(table_name, file_name)

    @staticmethod
    def get_file_name_from_file_path(file_path: str) -> str:
        return Path(file_path).name

    @staticmethod
    def make_job_with_status(file_path: str, status: LoadJobStatus, message: str = None) -> LoadJob:
        return LoadEmptyJob(ClientBase.get_file_name_from_file_path(file_path), status, exception=message)

    @staticmethod
    def make_absolute_path(file_path: str) -> str:
        return str(Path(file_path).absolute())


class SqlClientBase(ClientBase):
    def __init__(self, schema: Schema) -> None:
        super().__init__(schema)

    def complete_load(self, load_id: str) -> None:
        name = self._to_canonical_table_name(Schema.LOADS_TABLE_NAME)
        now_ts = str(pendulum.now())
        self._execute_sql(f"INSERT INTO {name}(load_id, status, inserted_at) VALUES('{load_id}', 0, '{now_ts}');")

    @abstractmethod
    def _execute_sql(self, query: AnyStr) -> Any:
        pass

    @abstractmethod
    def _to_canonical_schema_name(self) -> str:
        pass

    def _create_table_update(self, table_name: str, storage_table: Table) -> Sequence[Column]:
        # compare table with stored schema and produce delta
        l = self.schema.get_schema_update_for(table_name, storage_table)
        logger.info(f"Found {len(l)} updates for {table_name} in {self.schema.schema_name}")
        return l

    def _to_canonical_table_name(self, table_name: str) -> str:
        return f"{self._to_canonical_schema_name()}.{table_name}"

    def _get_schema_version_from_storage(self) -> int:
        name = self._to_canonical_table_name(Schema.VERSION_TABLE_NAME)
        rows = list(self._execute_sql(f"SELECT {Schema.VERSION_COLUMN_NAME} FROM {name} ORDER BY inserted_at DESC LIMIT 1;"))
        if len(rows) > 1:
            raise LoadClientSchemaVersionCorrupted(self._to_canonical_schema_name())
        if len(rows) == 0:
            return 0
        return int(rows[0][0])

    def _update_schema_version(self, new_version: int) -> None:
        now_ts = str(pendulum.now())
        name = self._to_canonical_table_name(Schema.VERSION_TABLE_NAME)
        self._execute_sql(f"INSERT INTO {name}({Schema.VERSION_COLUMN_NAME}, engine_version, inserted_at) VALUES ({new_version}, {Schema.ENGINE_VERSION}, '{now_ts}');")



from types import ModuleType
from typing import Any, Iterator, List, Dict, Literal, Optional, Tuple, Type
from multiprocessing.pool import ThreadPool
from importlib import import_module
from prometheus_client import REGISTRY, Counter, Gauge, CollectorRegistry, Summary
from prometheus_client.metrics import MetricWrapperBase

from dlt.common import sleep, logger
from dlt.common.runners import TRunArgs, TRunMetrics, create_default_args, initialize_runner, pool_runner
from dlt.common.logger import process_internal_exception, pretty_format_exception
from dlt.common.exceptions import TerminalValueError
from dlt.common.dataset_writers import TWriterType
from dlt.common.schema import Schema
from dlt.common.storages import SchemaStorage
from dlt.common.storages.loader_storage import LoaderStorage
from dlt.common.telemetry import get_logging_extras, set_gauge_all_labels

from dlt.loaders.exceptions import LoadClientTerminalException, LoadClientTransientException, LoadJobNotExistsException
from dlt.loaders.client_base import ClientBase, LoadJob
from dlt.loaders.local_types import LoadJobStatus
from dlt.loaders.configuration import configuration, LoaderConfiguration


CONFIG: Type[LoaderConfiguration] = None
load_storage: LoaderStorage = None
client_module: ModuleType = None
load_counter: Counter = None
job_gauge: Gauge = None
job_counter: Counter = None
job_wait_summary: Summary = None


def client_impl(client_type: str) -> ModuleType:
    return import_module(f".{client_type}.client", "dlt.loaders")


def create_client(schema: Schema) -> ClientBase:
    return client_module.make_client(schema, CONFIG)  # type: ignore


def supported_writer() -> TWriterType:
    return client_module.supported_writer(CONFIG)  # type: ignore


def create_folders(is_storage_owner: bool) -> LoaderStorage:
    load_storage = LoaderStorage(is_storage_owner, CONFIG, supported_writer())
    load_storage.initialize_storage()
    return load_storage


def create_gauges(registry: CollectorRegistry) -> Tuple[MetricWrapperBase, MetricWrapperBase, MetricWrapperBase, MetricWrapperBase]:
    return (
        Counter("loader_load_package_counter", "Counts load package processed", registry=registry),
        Gauge("loader_last_package_jobs_counter", "Counts jobs in last package per status", ["status"], registry=registry),
        Counter("loader_jobs_counter", "Counts jobs per job status", ["status"], registry=registry),
        Summary("loader_jobs_wait_seconds", "Counts jobs total wait until completion", registry=registry)
    )


def spool_job(file_path: str, load_id: str, schema: Schema) -> Optional[LoadJob]:
    # open new connection for each upload
    job: LoadJob = None
    try:
        with create_client(schema) as client:
            table_name, _ = load_storage.parse_load_file_name(file_path)
            logger.info(f"Will load file {file_path} with table name {table_name}")
            job = client.start_file_load(table_name, load_storage.storage._make_path(file_path))
    except (LoadClientTerminalException, TerminalValueError):
        # if job irreversible cannot be started, mark it as failed
        process_internal_exception(f"Terminal problem with spooling job {file_path}")
        job = ClientBase.make_job_with_status(file_path, "failed", pretty_format_exception())
    except (LoadClientTransientException, Exception):
        # return no job so file stays in new jobs (root) folder
        process_internal_exception(f"Temporary problem with spooling job {file_path}")
        return None
    load_storage.start_job(load_id, job.file_name())
    return job


def spool_new_jobs(pool: ThreadPool, load_id: str, schema: Schema) -> Tuple[int, List[LoadJob]]:
    # TODO: validate file type, combine files, finalize etc., this is client specific, jsonl for single table
    # can just be combined, insert_values must be finalized and then combined
    # use thread based pool as jobs processing is mostly I/O and we do not want to pickle jobs
    # TODO: combine files by providing a list of files pertaining to same table into job, so job must be
    # extended to accept a list
    load_files = load_storage.list_new_jobs(load_id)[:CONFIG.MAX_PARALLEL_LOADS]
    file_count = len(load_files)
    if file_count == 0:
        logger.info(f"No new jobs found in {load_id}")
        return 0, []
    logger.info(f"Will load {file_count}, creating jobs")
    param_chunk = [(file, load_id, schema) for file in load_files]
    # exceptions should not be raised, None as job is a temporary failure
    # other jobs should not be affected
    jobs: List[LoadJob] = pool.starmap(spool_job, param_chunk)
    # remove None jobs and check the rest
    return file_count, [job for job in jobs if job is not None]


def retrieve_jobs(client: ClientBase, load_id: str) -> Tuple[int, List[LoadJob]]:
    jobs: List[LoadJob] = []

    # list all files that were started but not yet completed
    started_jobs = load_storage.list_started_jobs(load_id)
    logger.info(f"Found {len(started_jobs)} that are already started and should be continued")
    if len(started_jobs) == 0:
        return 0, jobs

    for file_path in started_jobs:
        try:
            logger.info(f"Will retrieve {file_path}")
            job = client.get_file_load(file_path)
        except LoadClientTerminalException:
            process_internal_exception(f"Job retrieval for {file_path} failed, job will be terminated")
            job = ClientBase.make_job_with_status(file_path, "failed", pretty_format_exception())
            # proceed to appending job, do not reraise
        except (LoadClientTransientException, Exception) as e:
            # raise on all temporary exceptions, typically network / server problems
            raise
        jobs.append(job)

    job_gauge.labels("retrieved").inc()
    job_counter.labels("retrieved").inc()
    logger.metrics("Retrieve jobs metrics",
                    extra=get_logging_extras([job_gauge.labels("retrieved"), job_counter.labels("retrieved")])
    )
    return len(jobs), jobs


def complete_jobs(load_id: str, jobs: List[LoadJob]) -> List[LoadJob]:
    remaining_jobs: List[LoadJob] = []
    logger.info(f"Will complete {len(jobs)} for {load_id}")
    for ii in range(len(jobs)):
        job = jobs[ii]
        logger.debug(f"Checking status for job {job.file_name()}")
        status: LoadJobStatus = job.status()
        final_location: str = None
        if status == "running":
            # ask again
            logger.debug(f"job {job.file_name()} still running")
            remaining_jobs.append(job)
        elif status == "failed":
            # try to get exception message from job
            failed_message = job.exception()
            final_location = load_storage.fail_job(load_id, job.file_name(), failed_message)
            logger.error(f"Job for {job.file_name()} failed terminally in load {load_id} with message {failed_message}")
        elif status == "retry":
            # try to get exception message from job
            retry_message = job.exception()
            # move back to new folder to try again
            final_location = load_storage.retry_job(load_id, job.file_name())
            logger.error(f"Job for {job.file_name()} retried in load {load_id} with message {retry_message}")
        elif status == "completed":
            # move to completed folder
            final_location = load_storage.complete_job(load_id, job.file_name())
            logger.info(f"Job for {job.file_name()} completed in load {load_id}")

        if status != "running":
            job_gauge.labels(status).inc()
            job_counter.labels(status).inc()
            job_wait_summary.observe(load_storage.job_elapsed_time_seconds(final_location))

    logger.metrics("Completing jobs metrics", extra=get_logging_extras([job_counter, job_gauge, job_wait_summary]))
    return remaining_jobs



def load(pool: ThreadPool) -> TRunMetrics:
    logger.info(f"Running file loading")
    # get list of loads and order by name ASC to execute schema updates
    loads = load_storage.list_loads()
    logger.info(f"Found {len(loads)} load packages")
    if len(loads) == 0:
        return TRunMetrics(True, False, 0)

    load_id = loads[0]
    logger.info(f"Loading schema from load package in {load_id}")
    # one load package contains table from one schema
    schema_storage = SchemaStorage(load_storage.storage.storage_path)
    # get relative path to load schema from load package
    schema = schema_storage.load_folder_schema(load_storage.get_load_path(load_id))
    logger.info(f"Loaded schema name {schema.schema_name} and version {schema.schema_version}")
    # initialize analytical storage ie. create dataset required by passed schema
    with create_client(schema) as client:
        logger.info(f"Client {CONFIG.CLIENT_TYPE} will start load")
        client.initialize_storage()
        schema_update = load_storage.begin_schema_update(load_id)
        if schema_update:
            logger.info(f"Client {CONFIG.CLIENT_TYPE} will update schema to package schema")
            client.update_storage_schema()
            load_storage.commit_schema_update(load_id)
        # spool or retrieve unfinished jobs
        jobs_count, jobs = retrieve_jobs(client, load_id)
    if not jobs:
        # jobs count is a total number of jobs including those that could not be initialized
        jobs_count, jobs = spool_new_jobs(pool, load_id, schema)
        if jobs_count > 0:
            # this is a new  load package
            set_gauge_all_labels(job_gauge, 0)
            job_gauge.labels("running").inc(len(jobs))
            job_counter.labels("running").inc(len(jobs))
            logger.metrics("New jobs metrics",
                            extra=get_logging_extras([job_counter.labels("running"), job_gauge.labels("running")])
        )
    # if there are no existing or new jobs we archive the package
    if jobs_count == 0:
        with create_client(schema) as client:
            remaining_jobs = client.complete_load(load_id)
        load_storage.archive_load(load_id)
        logger.info(f"All jobs completed, archiving package {load_id}")
        load_counter.inc()
        logger.metrics("Load package metrics", extra=get_logging_extras([load_counter]))
    else:
        while True:
            remaining_jobs = complete_jobs(load_id, jobs)
            if len(remaining_jobs) == 0:
                break
            # process remaining jobs again
            jobs = remaining_jobs
            # this will raise on signal
            sleep(1)

    return TRunMetrics(False, False, len(load_storage.list_loads()))


def configure(C: Type[LoaderConfiguration], collector: CollectorRegistry, is_storage_owner: bool = False) -> None:
    global CONFIG
    global client_module, load_storage
    global load_counter, job_gauge, job_counter, job_wait_summary

    CONFIG = C
    client_module = client_impl(C.CLIENT_TYPE)
    load_storage = create_folders(is_storage_owner)
    try:
        load_counter, job_gauge, job_counter, job_wait_summary = create_gauges(collector)
    except ValueError as v:
        # ignore re-creation of gauges
        if "Duplicated timeseries" not in str(v):
            raise



def main(args: TRunArgs) -> int:
    C = configuration()
    initialize_runner(C, args)
    try:
        configure(C, REGISTRY)
    except Exception:
        process_internal_exception("run")
        return -1
    return pool_runner(C, load)


def run_main(args: TRunArgs) -> None:
    exit(main(args))



import argparse
from typing import Callable

from dlt.common.runners import TRunArgs, add_pool_cli_arguments


def main() -> None:
    parser = argparse.ArgumentParser(description="Runs various DLT modules")
    subparsers = parser.add_subparsers(dest="command")
    unpack = subparsers.add_parser("unpack", help="Runs unpacker")
    add_pool_cli_arguments(unpack)
    load = subparsers.add_parser("load", help="Runs loader")
    add_pool_cli_arguments(load)
    dbt = subparsers.add_parser("dbt", help="Executes dbt package")
    add_pool_cli_arguments(dbt)

    args = parser.parse_args()
    run_f: Callable[[TRunArgs], None] = None

    if args.command == "unpack":
        from dlt.unpacker.unpacker import run_main as unpacker_run
        run_f = unpacker_run
    elif args.command == "load":
        from dlt.loaders.loader import run_main as loader_run
        run_f = loader_run
    elif args.command == "dbt":
        from dlt.dbt_runner.runner import run_main as dbt_run
        run_f = dbt_run
    else:
        raise ValueError(args.command)

    run_args = TRunArgs(args.single_run, args.wait_runs)
    run_f(run_args)

if __name__ == "__main__":
    main()



import re

RE_UNDERSCORES = re.compile("_+")
RE_DOUBLE_UNDERSCORES = re.compile("__+")
RE_LEADING_DIGITS = re.compile(r"^\d+")
INVALID_SQL_IDENT_CHARS = "- *!:,.'\\\"`"
INVALID_SQL_TX = str.maketrans(INVALID_SQL_IDENT_CHARS, "_" * len(INVALID_SQL_IDENT_CHARS))
SNAKE_CASE_BREAK_1 = re.compile("(.)([A-Z][a-z]+)")
SNAKE_CASE_BREAK_2 = re.compile("([a-z0-9])([A-Z])")


# fix a name so it's acceptable as database table name
def normalize_table_name(name: str) -> str:

    def camel_to_snake(name: str) -> str:
        name = SNAKE_CASE_BREAK_1.sub(r'\1_\2', name)
        return SNAKE_CASE_BREAK_2.sub(r'\1_\2', name).lower()

    # all characters that are not letters digits or a few special chars are replaced with underscore
    # then convert to snake case
    name = camel_to_snake(name.translate(INVALID_SQL_TX))
    # leading digits will be removed
    name = RE_LEADING_DIGITS.sub("_", name)
    # max 2 consecutive underscores are allowed
    return RE_DOUBLE_UNDERSCORES.sub("__", name)

# fix a name so it's an acceptable name for a database column name
def normalize_db_name(name: str) -> str:
    # replace consecutive underscores with single one to prevent name clashes with parent child
    return RE_UNDERSCORES.sub("_", normalize_table_name(name))


def normalize_schema_name(name: str) -> str:
    # remove underscores
    return normalize_db_name(name).replace("_", "")



import sys
import semver
from os import environ
from os.path import isdir, isfile
from typing import Any, Dict, List, Mapping, NewType, Optional, Type, TypeVar, Union, Literal, IO, cast

from dlt.common.typing import StrAny
from dlt.common.configuration import BasicConfiguration
from dlt.common.configuration.exceptions import (ConfigEntryMissingException,
                                                         ConfigEnvValueCannotBeCoercedException, ConfigFileNotFoundException)
from dlt.common.utils import uniq_id

SIMPLE_TYPES: List[Any] = [int, bool, list, dict, tuple, bytes, set, float]
# those types and Optionals of those types should not be passed to eval function
NON_EVAL_TYPES = [str, None, Any]
# allows to coerce (type1 from type2)
ALLOWED_TYPE_COERCIONS = [(float, int), (str, int), (str, float)]
IS_DEVELOPMENT_CONFIG_KEY: str = "IS_DEVELOPMENT_CONFIG"
CHECK_INTEGRITY_F: str = "check_integrity"
SECRET_STORAGE_PATH: str = "/run/secrets/%s"

TConfiguration = TypeVar("TConfiguration", bound=Type[BasicConfiguration])
TProductionConfiguration = TypeVar("TProductionConfiguration", bound=Type[BasicConfiguration])
TConfigSecret = NewType("TConfigSecret", str)


def make_configuration(config: TConfiguration,
                       production_config: TProductionConfiguration,
                       initial_values: StrAny = None,
                       accept_partial: bool = False,
                       skip_subclass_check: bool = False) -> TConfiguration:
    if not skip_subclass_check:
        assert issubclass(production_config, config)

    final_config: TConfiguration = config if _is_development_config() else production_config
    possible_keys_in_config = _get_config_attrs_with_hints(final_config)
    # create dynamic class type to not touch original config variables
    derived_config: TConfiguration = cast(TConfiguration,
                                          type(final_config.__name__ + "_" + uniq_id(), (final_config, ), {})
                                    )
    # apply initial values while preserving hints
    if initial_values:
        for k, v in initial_values.items():
            setattr(derived_config, k, v)

    _apply_environ_to_config(derived_config, possible_keys_in_config)
    try:
        _is_config_bounded(derived_config, possible_keys_in_config)
        _check_configuration_integrity(derived_config)
    except ConfigEntryMissingException:
        if not accept_partial:
            raise
    _add_module_version(derived_config)

    return derived_config


def has_configuration_file(name: str, config: TConfiguration) -> bool:
    return isfile(get_configuration_file_path(name, config))


def open_configuration_file(name: str, mode: str, config: TConfiguration) -> IO[Any]:
    path = get_configuration_file_path(name, config)
    if not has_configuration_file(name, config):
        raise ConfigFileNotFoundException(path)
    return open(path, mode)


def get_configuration_file_path(name: str, config: TConfiguration) -> str:
    return config.CONFIG_FILES_STORAGE_PATH % name


def is_direct_descendant(child: Type[Any], base: Type[Any]) -> bool:
    # TODO: there may be faster way to get direct descendant that mro
    # note: at index zero there's child
    return base == type.mro(child)[1]


def _is_development_config() -> bool:
    is_dev_config = True

    # get from environment
    if IS_DEVELOPMENT_CONFIG_KEY in environ:
        is_dev_config = _coerce_single_value(IS_DEVELOPMENT_CONFIG_KEY, environ[IS_DEVELOPMENT_CONFIG_KEY], bool)
    return is_dev_config


def _add_module_version(config: TConfiguration) -> None:
    try:
        v = sys._getframe(1).f_back.f_globals["__version__"]
        semver.VersionInfo.parse(v)
        setattr(config, "_VERSION", v)  # noqa: B010
    except KeyError:
        pass


def _apply_environ_to_config(config: TConfiguration, keys_in_config: Mapping[str, type]) -> None:
    for key, hint in keys_in_config.items():
        value = _get_key_value(key, hint)
        if value is not None:
            value_from_environment_variable = _coerce_single_value(key, value, hint)
            # set value
            setattr(config, key, value_from_environment_variable)


def _get_key_value(key: str, hint: Type[Any]) -> Optional[str]:
    if hint is TConfigSecret:
        # try secret storage
        try:
            # must conform to RFC1123
            secret_name = key.lower().replace("_", "-")
            secret_path = SECRET_STORAGE_PATH % secret_name
            # kubernetes stores secrets as files in a dir, docker compose plainly
            if isdir(secret_path):
                secret_path += "/" + secret_name
            with open(secret_path, "r") as f:
                secret = f.read()
            # add secret to environ so forks have access
            # TODO: removing new lines is not always good. for password OK for PEMs not
            # TODO: in regular secrets that is dealt with in particular configuration logic
            environ[key] = secret.strip()
            # do not strip returned secret
            return secret

        except FileNotFoundError:
            pass
    return environ.get(key, None)


def _is_config_bounded(config: TConfiguration, keys_in_config: Mapping[str, type]) -> None:
    _unbound_attrs = [
        key for key in keys_in_config if getattr(config, key) is None and not _is_optional_type(keys_in_config[key])
    ]

    if len(_unbound_attrs) > 0:
        raise ConfigEntryMissingException(_unbound_attrs)


def _check_configuration_integrity(config: TConfiguration) -> None:
    # python multi-inheritance is cooperative and this would require that all configurations cooperatively
    # call each other check_integrity. this is not at all possible as we do not know which configs in the end will
    # be mixed together.

    # get base classes in order of derivation
    mro = type.mro(config)
    for c in mro:
        # check if this class implements check_integrity (skip pure inheritance to not do double work)
        if CHECK_INTEGRITY_F in c.__dict__ and callable(getattr(c, CHECK_INTEGRITY_F)):
            # access unbounded __func__ to pass right class type so we check settings of the tip of mro
            c.__dict__[CHECK_INTEGRITY_F].__func__(config)


def _coerce_single_value(key: str, value: str, hint: Type[Any]) -> Any:
    try:
        hint_primitive_type = _extract_simple_type(hint)
        if hint_primitive_type not in NON_EVAL_TYPES:
            # create primitive types out of strings
            typed_value = eval(value)  # nosec
            # for primitive types check coercion
            if hint_primitive_type in SIMPLE_TYPES and type(typed_value) != hint_primitive_type:
                # allow some exceptions
                coerce_exception = next(
                    (e for e in ALLOWED_TYPE_COERCIONS if e == (hint_primitive_type, type(typed_value))), None)
                if coerce_exception:
                    return hint_primitive_type(typed_value)
                else:
                    raise ConfigEnvValueCannotBeCoercedException(key, typed_value, hint)
            return typed_value
        else:
            return value
    except ConfigEnvValueCannotBeCoercedException:
        raise
    except Exception as exc:
        raise ConfigEnvValueCannotBeCoercedException(key, value, hint) from exc


def _extract_simple_type(hint: Type[Any]) -> Type[Any]:
    # extract optional type and call recursively
    if _is_literal_type(hint):
        # assume that all literals are of the same type
        return _extract_simple_type(type(hint.__args__[0]))
    if _is_optional_type(hint):
        # todo: use `get_args` in python 3.8
        return _extract_simple_type(hint.__args__[0])
    if not hasattr(hint, "__supertype__"):
        return hint
    # descend into supertypes of NewType
    return _extract_simple_type(hint.__supertype__)


def _get_config_attrs_with_hints(config: TConfiguration) -> Dict[str, type]:
    keys: Dict[str, type] = {}
    mro = type.mro(config)
    for cls in reversed(mro):
        # update in reverse derivation order so derived classes overwrite hints from base classes
        if cls is not object:
            keys.update(
                [(attr, cls.__annotations__.get(attr, None))
                  # if hasattr(config, '__annotations__') and attr in config.__annotations__ else None)
                 for attr in cls.__dict__.keys() if not callable(getattr(cls, attr)) and not attr.startswith("__")
                 ])
    return keys


def _is_optional_type(hint: Type[Any]) -> bool:
    # todo: use typing get_args and get_origin in python 3.8
    if hasattr(hint, "__origin__"):
        return hint.__origin__ is Union and type(None) in hint.__args__
    return False


def _is_literal_type(hint: Type[Any]) -> bool:
    return hasattr(hint, "__origin__") and hint.__origin__ is Literal




import base64
import binascii
import yaml
import datetime
import re
from re import Pattern
from copy import deepcopy
from dateutil.parser import isoparse
from typing import Dict, List, Set, Mapping, Optional, Sequence, Tuple, Type, TypedDict, Literal, Any, cast

from dlt.common import pendulum, json, Decimal
from dlt.common.names import normalize_schema_name
from dlt.common.typing import DictStrAny, StrAny, StrStr
from dlt.common.arithmetics import ConversionSyntax
from dlt.common.exceptions import DltException

DataType = Literal["text", "double", "bool", "timestamp", "bigint", "binary", "complex", "decimal", "wei"]
HintType = Literal["not_null", "partition", "cluster", "primary_key", "foreign_key", "sort", "unique"]
ColumnProp = Literal["name", "data_type", "nullable", "partition", "cluster", "primary_key", "foreign_key", "sort", "unique"]

DATA_TYPES: Set[DataType] = set(["text", "double", "bool", "timestamp", "bigint", "binary", "complex", "decimal", "wei"])
COLUMN_PROPS: Set[ColumnProp] = set(["name", "data_type", "nullable", "partition", "cluster", "primary_key", "foreign_key", "sort", "unique"])
COLUMN_HINTS: Set[HintType] = set(["partition", "cluster", "primary_key", "foreign_key", "sort", "unique"])

class ColumnBase(TypedDict, total=True):
    name: str
    data_type: DataType
    nullable: bool

class Column(ColumnBase, total=True):
    partition: bool
    cluster: bool
    unique: bool
    sort: bool
    primary_key: bool
    foreign_key: bool

Table = Dict[str, Column]
SchemaTables = Dict[str, Table]
SchemaUpdate = Dict[str, List[Column]]


class StoredSchema(TypedDict, total=True):
    version: int
    engine_version: int
    name: str
    tables: SchemaTables
    preferred_types: Mapping[str, DataType]
    hints: Mapping[HintType, Sequence[str]]
    excludes: Sequence[str]
    includes: Sequence[str]


class Schema:

    VERSION_TABLE_NAME = "_version"
    VERSION_COLUMN_NAME = "version"
    LOADS_TABLE_NAME = "_loads"
    ENGINE_VERSION = 2

    def __init__(self, name: str) -> None:
        # verify schema name
        if name != normalize_schema_name(name):
            raise InvalidSchemaName(name)
        self._schema_tables: SchemaTables = {}
        self._schema_name: str = name
        self._version = 1
        # list of preferred types: map regex on columns into types
        self._preferred_types: Mapping[str, DataType] = {}
        # compiled regexes
        self._compiled_preferred_types: List[Tuple[Pattern[str], DataType]] = []
        # table hints
        self._hints: Mapping[HintType, Sequence[str]] = {}
        self._compiled_hints: Dict[HintType, Sequence[Pattern[str]]] = {}
        # excluded paths
        self._excludes: Sequence[str] = []
        self._compiled_excludes: Sequence[Pattern[str]] = []
        # included paths
        self._includes: Sequence[str] = []
        self._compiled_includes: Sequence[Pattern[str]] = []
        # add version table
        self._add_standard_tables()
        # add standard hints
        self._add_standard_hints()
        # compile hints
        self._compile_regexes()

    @classmethod
    def from_dict(cls, stored_schema: StoredSchema) -> "Schema":
        # upgrade engine if needed
        cls._upgrade_engine_version(stored_schema, stored_schema["engine_version"], cls.ENGINE_VERSION)
        # create new instance from dict
        self: Schema = cls(stored_schema["name"])
        self._schema_tables = stored_schema["tables"]
        # TODO: generate difference if STANDARD SCHEMAS are different than those and increase schema version
        if Schema.VERSION_TABLE_NAME not in self._schema_tables:
            raise SchemaCorruptedException(f"Schema must contain table {Schema.VERSION_TABLE_NAME}")
        if Schema.LOADS_TABLE_NAME not in self._schema_tables:
            raise SchemaCorruptedException(f"Schema must contain table {Schema.LOADS_TABLE_NAME}")
        # verify table schemas
        for table_name, table in self._schema_tables.items():
            for column_name in table:
                # add default hints to tables
                column = self._add_missing_hints(table[column_name])
                # overwrite column name
                column["name"] = column_name
                # verify column
                self._verify_column(table_name, column_name, column)
                table[column_name] = column
        self._version = stored_schema["version"]
        self._preferred_types = stored_schema["preferred_types"]
        self._hints = stored_schema["hints"]
        self._excludes = stored_schema["excludes"]
        self._includes = stored_schema["includes"]
        # compile regexes
        self._compile_regexes()

        return self

    def filter_row(self, table_name: str, row: StrAny, path_separator: str) -> StrAny:
        # include and exclude paths follow the naming convention of the unpacker and correspond to json document nesting
        # current version of the unpacker separates json elements with __

        def _exclude(path: str) -> bool:
            is_included = False
            is_excluded = any(exclude.search(path) for exclude in self._compiled_excludes)
            if is_excluded:
                # we may have exception if explicitly included
                is_included = any(include.search(path) for include in self._compiled_includes)
            return is_excluded and not is_included

        # check if any of the rows is excluded
        for field_name in list(row.keys()):
            path = f"{table_name}{path_separator}{field_name}"
            # excluded if any rule matches
            if _exclude(path):
                # TODO: copy to new instance
                del row[field_name]  # type: ignore
        return row

    def coerce_row(self, table_name: str, row: StrAny) -> Tuple[StrAny, List[Column]]:
        table_schema: Table = self._schema_tables.get(table_name, {})
        new_columns: List[Column] = []
        new_row: DictStrAny = {}
        for col_name, v in row.items():
            # skip None values, we should infer the types later
            if v is None:
                # just check if column is nullable if exists
                self._coerce_null_value(table_schema, table_name, col_name)
            else:
                new_col_name, new_col_def, new_v = self._coerce_non_null_value(table_schema, table_name, col_name, v)
                new_row[new_col_name] = new_v
                if new_col_def:
                    new_columns.append(new_col_def)

        return new_row, new_columns

    def filter_hints_in_row(self, table_name: str, hint_type: HintType, row: StrAny) -> StrAny:
        rv_row: DictStrAny = {}
        column_prop: ColumnProp = self._hint_to_column_prop(hint_type)
        try:
            table = self.get_table(table_name)
            for column_name in table:
                if column_name in row:
                    hint_value = table[column_name][column_prop]
                    if (hint_value and column_prop != "nullable") or (column_prop == "nullable" and not hint_value):
                        rv_row[column_name] = row[column_name]
        except KeyError:
            for k, v in row.items():
                if self._infer_hint(hint_type, v, k):
                    rv_row[k] = v

        # dicts are ordered and we will return the rows with hints in the same order as they appear in the columns
        return rv_row

    def update_schema(self, table_name: str, updated_columns: List[Column]) -> None:
        # all tables in the schema must start with the schema name
        # if not table_name.startswith(f"{self._schema_name}"):
        #     raise InvalidTableNameException(self._schema_name, table_name)

        if table_name not in self._schema_tables:
            # add the whole new table to SchemaTables
            self._schema_tables[table_name] = {c["name"]: c for c in updated_columns}
        else:
            # add several columns to existing table
            table_schema = self._schema_tables[table_name]
            for column in updated_columns:
                column_name = column["name"]
                if column_name in table_schema:
                    # we do not support changing existing columns
                    if not Schema._compare_columns(table_schema[column_name], column):
                        # attempt to update to incompatible columns
                        raise CannotCoerceColumnException(table_name, column_name, table_schema[column_name]["data_type"], column["data_type"], None)
                else:
                    table_schema[column_name] = column
        # bump schema version
        self._version += 1

    def get_schema_update_for(self, table_name: str, t: Table) -> List[Column]:
        # gets new columns to be added to "t" to bring up to date with stored schema
        diff_c: List[Column] = []
        s_t = self.get_table(table_name)
        for c in s_t.values():
            if c["name"] not in t:
                diff_c.append(c)
        return diff_c

    def get_table(self, table_name: str) -> Table:
        return self._schema_tables[table_name]

    def to_dict(self) -> StoredSchema:
        return  {
            "tables": self._schema_tables,
            "name": self._schema_name,
            "version": self._version,
            "preferred_types": self._preferred_types,
            "hints": self._hints,
            "excludes": self._excludes,
            "includes": self._includes,
            "engine_version": Schema.ENGINE_VERSION
        }

    @property
    def schema_version(self) -> int:
        return self._version

    @property
    def schema_name(self) -> str:
        return self._schema_name

    @property
    def schema_tables(self) -> SchemaTables:
        return self._schema_tables

    def as_yaml(self, remove_default_hints: bool = False) -> str:
        d = self.to_dict()
        clean_tables = deepcopy(d["tables"])

        for t in clean_tables.values():
            for c in t.values():
                # do not save names
                del c["name"]  # type: ignore
                # remove hints with default values
                if remove_default_hints:
                    for h in list(c.keys()):
                        if isinstance(c[h], bool) and c[h] is False and h != "nullable":  # type: ignore
                            del c[h]  # type: ignore

        d["tables"] = clean_tables

        return cast(str, yaml.dump(d, allow_unicode=True, default_flow_style=False, sort_keys=False))

    def _infer_column(self, k: str, v: Any) -> Column:
        return Column(
            name=k,
            data_type=self._map_value_to_column_type(v, k),
            nullable=not self._infer_hint("not_null", v, k),
            partition=self._infer_hint("partition", v, k),
            cluster=self._infer_hint("cluster", v, k),
            sort=self._infer_hint("sort", v, k),
            unique=self._infer_hint("unique", v, k),
            primary_key=self._infer_hint("primary_key", v, k),
            foreign_key=self._infer_hint("foreign_key", v, k)
        )

    def _coerce_null_value(self, table_schema: Table, table_name: str, col_name: str) -> None:
        if col_name in table_schema:
            existing_column = table_schema[col_name]
            if not existing_column["nullable"]:
                raise CannotCoerceNullException(table_name, col_name)

    def _coerce_non_null_value(self, table_schema: Table, table_name: str, col_name: str, v: Any) -> Tuple[str, Column, Any]:
        new_column: Column = None
        rv = v
        variant_col_name = col_name

        if col_name in table_schema:
            existing_column = table_schema[col_name]
            # existing columns cannot be changed so we must update row
            py_data_type = Schema._py_type_to_sc_type(type(v))
            if existing_column["data_type"] != py_data_type:
                # first try to coerce existing value into destination type
                try:
                    rv = Schema._coerce_type(existing_column["data_type"], py_data_type, v)
                except (ValueError, SyntaxError):
                    # for complex types we must coerce to text
                    if py_data_type == "complex":
                        py_data_type = "text"
                        rv = Schema._coerce_type("text", "complex", v)
                    # if that does not work we must create variant extension to the table
                    variant_col_name = f"{col_name}_v_{py_data_type}"
                    # if variant exists check type, coercions are not required
                    if variant_col_name in table_schema:
                        if table_schema[variant_col_name]["data_type"] != py_data_type:
                            raise CannotCoerceColumnException(table_name, variant_col_name, table_schema[variant_col_name]["data_type"], py_data_type, v)
                    else:
                        # new column
                        # add new column
                        new_column = self._infer_column(variant_col_name, v)
                        # must have variant type, not preferred or coerced type
                        new_column["data_type"] = py_data_type
            else:
                # just copy row: types match
                pass
        else:
            # infer new column
            new_column = self._infer_column(col_name, v)
            # and coerce type if inference changed the python type
            py_type = Schema._py_type_to_sc_type(type(v))
            rv = Schema._coerce_type(new_column["data_type"], py_type, v)

        return variant_col_name, new_column, rv

    def _map_value_to_column_type(self, v: Any, k: str) -> DataType:
        mapped_type = Schema._py_type_to_sc_type(type(v))
         # if complex type was detected we must coerce to string
        if mapped_type == "complex":
            mapped_type = "text"
        # get preferred type based on column name
        preferred_type = self._get_preferred_type(k)
        # try to match python type to preferred
        if preferred_type:
            # try to coerce to destination type
            try:
                Schema._coerce_type(preferred_type, mapped_type, v)
                # coercion possible so preferred type may be used
                mapped_type = preferred_type
            except ValueError:
                # coercion not possible
                pass
        return mapped_type

    def _get_preferred_type(self, col_name: str) -> Optional[DataType]:
        return next((m[1] for m in self._compiled_preferred_types if m[0].search(col_name)), None)

    def _infer_hint(self, hint_type: HintType, _: Any, k: str) -> bool:
        if hint_type in self._compiled_hints:
            return any(h.search(k) for h in self._compiled_hints[hint_type])
        else:
            return False

    def _add_standard_tables(self) -> None:
        version_table: Table = {
            "version": self._add_missing_hints({
                "name": "version",
                "data_type": "bigint",
                "nullable": False,
            }),
            "engine_version": self._add_missing_hints({
                "name": "engine_version",
                "data_type": "bigint",
                "nullable": False
            }),
            "inserted_at": self._add_missing_hints({
                "name": "inserted_at",
                "data_type": "timestamp",
                "nullable": False
            })

        }
        self._schema_tables[Schema.VERSION_TABLE_NAME] = version_table
        load_table: Table = {
            "load_id": self._add_missing_hints({
                "name": "load_id",
                "data_type": "text",
                "nullable": False
            }),
            "status": self._add_missing_hints({
                "name": "status",
                "data_type": "bigint",
                "nullable": False
            }),
            "inserted_at": self._add_missing_hints({
                "name": "inserted_at",
                "data_type": "timestamp",
                "nullable": False
            })
        }
        self._schema_tables[Schema.LOADS_TABLE_NAME] = load_table

    def _add_standard_hints(self) -> None:
        self._hints = {
            "not_null": ["^_record_hash$", "^_root_hash$", "^_parent_hash$", "^_pos$", "_load_id"],
            "foreign_key": ["^_parent_hash$"],
            "unique": ["^_record_hash$"]
        }

    def _compile_regexes(self) -> None:
        for pattern, dt in self._preferred_types.items():
            # add tuples to be searched in coercions
            self._compiled_preferred_types.append((re.compile(pattern), dt))
        for hint_name, hint_list in self._hints.items():
            # compile hints which are column matching regexes
            self._compiled_hints[hint_name] = list(map(lambda hint: re.compile(hint), hint_list))
        self._compiled_excludes = list(map(lambda exclude: re.compile(exclude), self._excludes))
        self._compiled_includes = list(map(lambda include: re.compile(include), self._includes))

    @staticmethod
    def _verify_column(table_name: str, column_name: str, column: Column) -> None:
        existing_props = set(column.keys())
        missing_props = COLUMN_PROPS.difference(existing_props)
        if len(missing_props) > 0:
            raise SchemaCorruptedException(f"In table {table_name} column {column_name}: Column definition is missing following properties {missing_props}")
        data_type = column["data_type"]
        if data_type not in DATA_TYPES:
            raise SchemaCorruptedException(f"In table {table_name} column {column_name}: {data_type} is not one of available types: {DATA_TYPES}")
        for p, v in column.items():
            if p in COLUMN_HINTS and not isinstance(v, bool):
                 raise SchemaCorruptedException(f"In table {table_name} column {column_name}: hint {p} is not boolean.")

    @staticmethod
    def _upgrade_engine_version(schema_dict: StoredSchema, from_engine: int, to_engine: int) -> None:
        if from_engine == 1:
            schema_dict["engine_version"] = 2
            schema_dict["includes"] = []
            schema_dict["excludes"] = []
            from_engine = 2
        if from_engine == 2:
            pass
        if from_engine != to_engine:
            raise SchemaEngineNoUpgradePathException(schema_dict["name"], schema_dict["engine_version"], from_engine, to_engine)

    @staticmethod
    def _add_missing_hints(column: ColumnBase) -> Column:
        return {
            **{  # type:ignore
                "partition": False,
                "cluster": False,
                "unique": False,
                "sort": False,
                "primary_key": False,
                "foreign_key": False,
            },
            **column
        }


    @staticmethod
    def _py_type_to_sc_type(t: Type[Any]) -> DataType:
        if t is float:
            return "double"
        elif t is int:
            return "bigint"
        elif t is bool:
            return "bool"
        elif issubclass(t, bytes):
            return "binary"
        elif issubclass(t, dict) or  issubclass(t, list):
            return "complex"
        elif issubclass(t, Decimal):
            return "decimal"
        elif issubclass(t, datetime.datetime):
            return "timestamp"
        else:
            return "text"

    @staticmethod
    def _coerce_type(to_type: DataType, from_type: DataType, value: Any) -> Any:
        if to_type == from_type:
            return value

        if to_type == "text":
            if from_type == "complex":
                return json.dumps(value)
            else:
                return str(value)

        if to_type == "binary":
            if from_type == "text":
                if value.startswith("0x"):
                    return bytes.fromhex(value[2:])
                try:
                    return base64.b64decode(value, validate=True)
                except binascii.Error:
                    raise ValueError(value)
            if from_type == "bigint":
                return value.to_bytes((value.bit_length() + 7) // 8, 'little')

        if to_type in ["wei", "bigint"]:
            if from_type == "bigint":
                return value
            if from_type in ["decimal", "double"]:
                if value % 1 != 0:
                    # only integer decimals and floats can be coerced
                    raise ValueError(value)
                return int(value)
            if from_type == "text":
                trim_value = value.strip()
                if trim_value.startswith("0x"):
                    return int(trim_value[2:], 16)
                else:
                    return int(trim_value)

        if to_type == "double":
             if from_type in ["bigint", "wei", "decimal"]:
                return float(value)
             if from_type == "text":
                trim_value = value.strip()
                if trim_value.startswith("0x"):
                    return float(int(trim_value[2:], 16))
                else:
                    return float(trim_value)

        if to_type == "decimal":
            if from_type in ["bigint", "wei"]:
                return value
            if from_type == "double":
                return Decimal(value)
            if from_type == "text":
                trim_value = value.strip()
                if trim_value.startswith("0x"):
                    return int(trim_value[2:], 16)
                elif "." not in trim_value and "e" not in trim_value:
                    return int(trim_value)
                else:
                    try:
                        return Decimal(trim_value)
                    except ConversionSyntax:
                        raise ValueError(trim_value)

        if to_type == "timestamp":
            if from_type in ["bigint", "double"]:
                # returns ISO datetime with timezone
                return str(pendulum.from_timestamp(value))

            if from_type == "text":
                # if parses as ISO date then pass it
                try:
                    isoparse(value)
                    return value
                except ValueError:
                    # try to convert string to integer, or float
                    try:
                        value = int(value)
                    except ValueError:
                        # raises ValueError if not parsing correctly
                        value = float(value)
                    return str(pendulum.from_timestamp(value))

        raise ValueError(value)

    @staticmethod
    def _compare_columns(a: Column, b: Column) -> bool:
        return a["data_type"] == b["data_type"] and a["nullable"] == b["nullable"]

    @staticmethod
    def _hint_to_column_prop(h: HintType) -> ColumnProp:
        if h == "not_null":
            return "nullable"
        return h

class SchemaException(DltException):
    pass


class InvalidSchemaName(SchemaException):
    def __init__(self, name: str) -> None:
        self.name = name
        super().__init__(f"{name} is invalid schema name. Only lowercase letters are allowed. Try {normalize_schema_name(name)} instead")


class CannotCoerceColumnException(SchemaException):
    def __init__(self, table_name: str, column_name: str, from_type: DataType, to_type: DataType, value: Any) -> None:
        super().__init__(f"Cannot coerce type in table {table_name} column {column_name} existing type {from_type} coerced type {to_type} value: {value}")


class CannotCoerceNullException(SchemaException):
    def __init__(self, table_name: str, column_name: str) -> None:
        super().__init__(f"Cannot coerce NULL in table {table_name} column {column_name} which is not nullable")


class InvalidTableNameException(SchemaException):
    def __init__(self, schema_name: str, table_name: str) -> None:
        self.schema_name = schema_name
        self.table_name = table_name
        super().__init__(f"All table names must start with '{schema_name}' so {table_name} is invalid")

class SchemaCorruptedException(SchemaException):
    pass


class SchemaEngineNoUpgradePathException(SchemaException):
    def __init__(self, schema_name: str, init_engine: int, from_engine: int, to_engine: int) -> None:
        self.schema_name = schema_name
        self.init_engine = init_engine
        self.from_engine = from_engine
        self.to_engine = to_engine
        super().__init__(f"No engine upgrade path in schema {schema_name} from {init_engine} to {to_engine}, stopped at {from_engine}")



from .pendulum import pendulum  # noqa: F401
from .json import json  # noqa: F401, I251
from .time import sleep   # noqa: F401
from .arithmetics import Decimal  # noqa: F401
from dlt._version import common_version as __version__


from typing import Iterator, Optional, Tuple, Callable, cast, TypedDict, Any

from dlt.common import json
from dlt.common.schema import Schema
from dlt.common.utils import uniq_id, digest128
from dlt.common.typing import TEvent, StrAny
from dlt.common.names import normalize_db_name
from dlt.common.sources import DLT_METADATA_FIELD, TEventDLTMeta, get_table_name


class TEventRow(TypedDict, total=False):
    _timestamp: float  # used for partitioning
    _dist_key: str  # distribution key used for clustering
    _record_hash: str  # unique id of current row
    _root_hash: str  # unique id of top level parent


class TEventRowRoot(TEventRow, total=False):
    _load_id: str  # load id to identify records loaded together that ie. need to be processed
    _event_json: str  # dump of the original event
    __dlt_meta: TEventDLTMeta  # stores metadata


class TEventRowChild(TEventRow, total=False):
    _parent_hash: str  # unique id of parent row
    _pos: int  # position in the list of rows
    value: Any  # for lists of simple types


# I(table name, row data)
TUnpackedRowIterator = Iterator[Tuple[str, StrAny]]
TExtractFunc = Callable[[Schema, TEvent, str, bool], TUnpackedRowIterator]


# subsequent nested fields will be separated with the string below, applies both to field and table names
PATH_SEPARATOR = "__"


# for those paths the complex nested objects should be left in place
# current use case: we want to preserve event_slot__value in db even if it's an object
# TODO: pass table definition and accept complex type
def _should_preserve_complex_value(table: str, field_name: str) -> bool:
    path = f"{table}{PATH_SEPARATOR}{field_name}"
    return path in ["event_slot__value"]


def _flatten(table: str, dict_row: TEventRowChild) -> TEventRowChild:
    out_rec_row: TEventRowChild = {}

    def unpack_row_dicts(dict_row: StrAny, parent_name: Optional[str]) -> None:
        for k, v in dict_row.items():
            corrected_k = normalize_db_name(k)
            child_name = corrected_k if not parent_name else f'{parent_name}{PATH_SEPARATOR}{corrected_k}'
            if isinstance(v, dict):
                unpack_row_dicts(v, parent_name=child_name)
                if _should_preserve_complex_value(table, child_name):
                    out_rec_row[child_name] = v  # type: ignore
            else:
                out_rec_row[child_name] = v  # type: ignore

    unpack_row_dicts(dict_row, None)
    return out_rec_row


def _get_child_row_hash(parent_hash: str, child_table: str, list_pos: int) -> str:
    # create deterministic unique id of the child row taking into account that all lists are ordered
    # and all child tables must be lists
    return digest128(f"{parent_hash}_{child_table}_{list_pos}")


def _unpack_row(
    schema: Schema,
    dict_row: TEventRowChild,
    extend: TEventRowChild,
    table: str,
    parent_hash: Optional[str] = None,
    pos: Optional[int] = None
    ) -> TUnpackedRowIterator:

    def _append_child_meta(_row: TEventRowChild, _hash: str, _p_hash: str, _p_pos: int) -> TEventRowChild:
        _row["_parent_hash"] = _p_hash
        _row["_pos"] = _p_pos
        _row.update(extend)

        return _row

    is_top_level = parent_hash is None

    # flatten current row
    new_dict_row = _flatten(table, dict_row)
    # infer record hash or leave existing primary key if present
    record_hash = new_dict_row.get("_record_hash", None)
    if not record_hash:
        # check if we have primary key: if so use it
        primary_key = schema.filter_hints_in_row(table, "primary_key", new_dict_row)
        if primary_key:
            # create row id from primary key
            record_hash = digest128("_".join(map(lambda v: str(v), primary_key.values())))
        elif not is_top_level:
            # child table row deterministic hash
            record_hash = _get_child_row_hash(parent_hash, table, pos)
            # link to parent table
            _append_child_meta(new_dict_row, record_hash, parent_hash, pos)
        else:
            # create random row id, note that incremental loads will not work with such tables
            record_hash = uniq_id()
        new_dict_row["_record_hash"] = record_hash

    # if _root_hash propagation requested and we are at the top level then update extend
    if "_root_hash" in extend and extend["_root_hash"] is None and is_top_level:
        extend["_root_hash"] = record_hash

    # generate child tables only for lists
    children = [k for k in new_dict_row if isinstance(new_dict_row[k], list)]  # type: ignore
    for k in children:
        child_table = f"{table}{PATH_SEPARATOR}{k}"
        # this will skip empty lists
        v: TEventRowChild
        for idx, v in enumerate(new_dict_row[k]):  # type: ignore
            # yield child table row
            tv = type(v)
            if tv is dict:
                yield from _unpack_row(schema, v, extend, child_table, record_hash, idx)
            elif tv is list:
                # unpack lists of lists
                raise ValueError(v)
            else:
                # list of simple types
                child_row_hash = _get_child_row_hash(record_hash, child_table, idx)
                e = _append_child_meta({"value": v, "_record_hash": child_row_hash}, child_row_hash, record_hash, idx)
                yield child_table, e
        if not _should_preserve_complex_value(table, k):
            # remove child list
            del new_dict_row[k]  # type: ignore

    yield table, new_dict_row


def extract(schema: Schema, source_event: TEvent, load_id: str, add_json: bool) -> TUnpackedRowIterator:
    # we will extend event with all the fields necessary to load it as root row
    event = cast(TEventRowRoot, source_event)
    # identify load id if loaded data must be processed after loading incrementally
    event["_load_id"] = load_id
    # add original json field, mostly useful for debugging
    if add_json:
        event["_event_json"] = json.dumps(event)
    # find table name
    table_name = get_table_name(event) or schema.schema_name
    # drop dlt metadata before unpacking
    event.pop(DLT_METADATA_FIELD, None)  # type: ignore
    # TODO: if table_name exist get "_dist_key" and "_timestamp" from the table definition in schema and propagate, if not take them from global hints
    # use event type or schema name as table name, request _root_hash propagation
    yield from _unpack_row(schema, cast(TEventRowChild, event), {"_root_hash": None}, table_name)



from typing import Any, Type

from dlt.common.utils import uniq_id
from dlt.common.typing import StrAny
from dlt.common.configuration import (PoolRunnerConfiguration,
                                              LoadingVolumeConfiguration,
                                              ProductionLoadingVolumeConfiguration,
                                              PostgresConfiguration, PostgresProductionConfiguration,
                                              GcpClientConfiguration, GcpClientProductionConfiguration,
                                              TPoolType, make_configuration)

from dlt.loaders.dummy.configuration import DummyClientConfiguration

from . import __version__

class LoaderConfiguration(PoolRunnerConfiguration, LoadingVolumeConfiguration):
    CLIENT_TYPE: str = "dummy"  # which analytical storage to use
    MAX_PARALLEL_LOADS: int = 20  # how many parallel loads can be executed
    MAX_PARALLELISM: int = 20  # in 20 separate threads
    POOL_TYPE: TPoolType = "thread"  # mostly i/o (upload) so may be thread pool


class ProductionLoaderConfiguration(ProductionLoadingVolumeConfiguration, LoaderConfiguration):
    pass


def configuration(initial_values: StrAny = None) -> Type[LoaderConfiguration]:
    # synthesize right configuration
    C = make_configuration(LoaderConfiguration, ProductionLoaderConfiguration, initial_values=initial_values)
    T: Type[Any] = None
    T_P: Type[Any] = None
    if C.CLIENT_TYPE == "dummy":
        T = DummyClientConfiguration
        T_P = DummyClientConfiguration
    elif C.CLIENT_TYPE == "gcp":
        T = GcpClientConfiguration
        T_P = GcpClientProductionConfiguration
    elif C.CLIENT_TYPE == "redshift":
        T = PostgresConfiguration
        T_P = PostgresProductionConfiguration
    else:
        raise ValueError(C.CLIENT_TYPE)

    ST = type(LoaderConfiguration.__name__ + "_"  + T.__name__ + "_" + uniq_id(), (T, LoaderConfiguration), {})
    ST_P = type(ProductionLoaderConfiguration.__name__ + "_" + T_P.__name__ + "_" + uniq_id(), (T_P, ProductionLoaderConfiguration), {})
    return make_configuration(
        ST,
        ST_P,
        initial_values=initial_values,
        skip_subclass_check=True
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dlt',
 'dlt.cli',
 'dlt.common',
 'dlt.common.configuration',
 'dlt.common.storages',
 'dlt.dbt_runner',
 'dlt.extractors',
 'dlt.extractors.generator',
 'dlt.loaders',
 'dlt.loaders.dummy',
 'dlt.loaders.gcp',
 'dlt.loaders.redshift',
 'dlt.pipeline',
 'dlt.unpacker',
 'examples',
 'examples.schemas',
 'examples.sources']

package_data = \
{'': ['*'], 'examples': ['data/*']}

install_requires = \
['PyYAML>=5.4.1,<6.0.0',
 'cachetools>=5.2.0,<6.0.0',
 'hexbytes>=0.2.2,<0.3.0',
 'json-logging==1.4.1rc0',
 'jsonlines>=2.0.0,<3.0.0',
 'pendulum>=2.1.2,<3.0.0',
 'prometheus-client>=0.11.0,<0.12.0',
 'requests>=2.26.0,<3.0.0',
 'semver>=2.13.0,<3.0.0',
 'sentry-sdk>=1.4.3,<2.0.0',
 'simplejson>=3.17.5,<4.0.0']

extras_require = \
{'dbt': ['GitPython>=3.1.26,<4.0.0',
         'dbt-core==1.0.6',
         'dbt-redshift==1.0.1',
         'dbt-bigquery==1.0.0'],
 'gcp': ['grpcio==1.43.0', 'google-cloud-bigquery>=2.26.0,<3.0.0'],
 'postgres': ['psycopg2-binary>=2.9.1,<3.0.0'],
 'redshift': ['psycopg2-binary>=2.9.1,<3.0.0']}

entry_points = \
{'console_scripts': ['dlt = dlt.cli.dlt:main']}

setup_kwargs = {
    'name': 'python-dlt',
    'version': '0.1.0rc3',
    'description': 'DLT is an open-source python-native scalable data loading framework that does not require any devops efforts to run.',
    'long_description': 'Follow this quick guide to implement DLT in your project\n\n## Simple loading of two rows:\n\n### Install DLT\nDLT is available in PyPi and can be installed with `pip install python-dlt`. Support for target warehouses is provided in extra packages:\n\n`pip install python-dlt[redshift]` for Redshift\n\n`pip install python-dlt[gcp]` for BigQuery\n\n\n### 1. Configuraton: name your schema, table, pass credentials\n\n```\n\nfrom dlt.pipeline import Pipeline, PostgresPipelineCredentials\n\nschema_prefix = \'demo_\'\nschema_name = \'example\'\nparent_table = \'my_json_doc\'\n\n# Credentials for Redshift or Bigquery\n\n# Redshift: you can pass password as last parameter or via PG_PASSWORD env variable.\n# credential = PostgresPipelineCredentials("redshift", "dbname", "schemaname", "username", "3.73.90.3", "dolphins")\n\n# Bigquery:\ngcp_credential_json_file_path = "scalevector-1235ac340b0b.json"\ncredential = Pipeline.load_gcp_credentials(gcp_credential_json_file_path, schema_prefix)\n\n# optionally save and reuse schema\n# schema_file_path = "examples/schemas/quickstart.yml"\n\n```\n\n### 2. Create a pipeline\n```\n\npipeline = Pipeline(schema_name)\npipeline.create_pipeline(credential)\n\n# Optionally uncomment to re-use a schema\n# schema = Pipeline.load_schema_from_file(schema_file_path)\n# pipeline.create_pipeline(credential, schema=schema)\n\n```\n\n### 3. Pass the data to the pipeline and give it a table name.\n```\n\nrows = [{"name":"Ana", "age":30, "id":456, "children":[{"name": "Bill", "id": 625},\n                                                       {"name": "Elli", "id": 591}\n                                                      ]},\n\n        {"name":"Bob", "age":30, "id":455, "children":[{"name": "Bill", "id": 625},\n                                                       {"name": "Dave", "id": 621}\n                                                      ]}\n       ]\n\npipeline.extract(iter(rows), table_name=parent_table)\n\n# Optionally the pipeline to un-nest the json into a relational structure\npipeline.unpack()\n\n# Optionally save the schema for manual edits/future use.\n# schema = pipeline.get_default_schema()\n# schema_yaml = schema.as_yaml()\n# f = open(data_schema_file_path, "a")\n# f.write(schema_yaml)\n# f.close()\n\n```\n\n### 4. Load\n\n```\npipeline.load()\n\n```\n\n### 5. Error capture - print, raise or handle.\n\n```\n# now enumerate all complete loads to check if we have any failed packages\n# complete but failed job will not raise any exceptions\ncompleted_loads = pipeline.list_completed_loads()\n# print(completed_loads)\nfor load_id in completed_loads:\n    print(f"Checking failed jobs in {load_id}")\n    for job, failed_message in pipeline.list_failed_jobs(load_id):\n        print(f"JOB: {job}\\nMSG: {failed_message}")\n\n```\n### 6. Use your data\n\n\nTables created:\n```\n SELECT *  FROM `scalevector.demo__example.my_json_doc`\nRESULT:\n{  "name": "Ana",  "age": "30",  "id": "456",  "_load_id": "1654787700.406905",  "_record_hash": "5b018c1ba3364279a0ca1a231fbd8d90"}\n{  "name": "Bob",  "age": "30",  "id": "455",  "_load_id": "1654787700.406905",  "_record_hash": "afc8506472a14a529bf3e6ebba3e0a9e"}\n\n\n SELECT * FROM `scalevector.demo__example.my_json_doc__children` LIMIT 1000\nRESULT:\n{  "name": "Bill",  "id": "625",  "_parent_hash": "5b018c1ba3364279a0ca1a231fbd8d90",  "_pos": "0",  "_root_hash": "5b018c1ba3364279a0ca1a231fbd8d90",  "_record_hash": "7993452627a98814cc7091f2c51faf5c"}\n{  "name": "Bill",  "id": "625",  "_parent_hash": "afc8506472a14a529bf3e6ebba3e0a9e",  "_pos": "0",  "_root_hash": "afc8506472a14a529bf3e6ebba3e0a9e",  "_record_hash": "9a2fd144227e70e3aa09467e2358f934"}\n{  "name": "Dave",  "id": "621",  "_parent_hash": "afc8506472a14a529bf3e6ebba3e0a9e",  "_pos": "1",  "_root_hash": "afc8506472a14a529bf3e6ebba3e0a9e",  "_record_hash": "28002ed6792470ea8caf2d6b6393b4f9"}\n{  "name": "Elli",  "id": "591",  "_parent_hash": "5b018c1ba3364279a0ca1a231fbd8d90",  "_pos": "1",  "_root_hash": "5b018c1ba3364279a0ca1a231fbd8d90",  "_record_hash": "d18172353fba1a492c739a7789a786cf"}\n\n```\nJoin your data via recursively created join keys.\n```\n select p.name, p.age, p.id as parent_id,\n        c.name as child_name, c.id as child_id, c._pos as child_order_in_list\n from `scalevector.demo__example.my_json_doc` as p\n left join `scalevector.demo__example.my_json_doc__children`  as c\n     on p._record_hash = c._parent_hash\nRESULT:\n{  "name": "Ana",  "age": "30",  "parent_id": "456",  "child_name": "Bill",  "child_id": "625",  "child_order_in_list": "0"}\n{  "name": "Ana",  "age": "30",  "parent_id": "456",  "child_name": "Elli",  "child_id": "591",  "child_order_in_list": "1"}\n{  "name": "Bob",  "age": "30",  "parent_id": "455",  "child_name": "Bill",  "child_id": "625",  "child_order_in_list": "0"}\n{  "name": "Bob",  "age": "30",  "parent_id": "455",  "child_name": "Dave",  "child_id": "621",  "child_order_in_list": "1"}\n\n```\n\n\n### 7. Run it yourself - plug your own iterator or generator.\nWorking example:\n[quickstart.py](examples/quickstart.py)',
    'author': 'ScaleVector',
    'author_email': 'services@scalevector.ai',
    'maintainer': 'Marcin Rudolf',
    'maintainer_email': 'marcin@scalevector.ai',
    'url': 'https://github.com/scale-vector',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



from apiclient.discovery import build
from oauth2client.service_account import ServiceAccountCredentials
from googleapiclient.http import MediaIoBaseDownload
import io
from typing import Any, Iterator
import csv

from dlt.common.typing import DictStrAny, StrAny
from dlt.common.schema import Schema

from dlt.pipeline import Pipeline, GCPPipelineCredentials

SCOPES = ['https://www.googleapis.com/auth/drive']
# KEY_FILE_LOCATION = '/Users/adrian/PycharmProjects/sv/dlt/temp/scalevector-1235ac340b0b.json'
KEY_FILE_LOCATION = '_secrets/scalevector-1235ac340b0b.json'


def _initialize_drive() -> Any:
    """Initializes an drive service object.

    Returns:
    An authorized drive service object.
    """
    credentials = ServiceAccountCredentials.from_json_keyfile_name(
        KEY_FILE_LOCATION, SCOPES)

    # Build the service object.
    service = build('drive', 'v3', credentials=credentials)

    return service


def download_csv_as_json(file_id: str, csv_options: StrAny = None) -> Iterator[DictStrAny]:
    if csv_options is None:
        csv_options = {}

    drive_service = _initialize_drive()
    request = drive_service.files().get_media(fileId=file_id)
    fh = io.BytesIO()
    downloader = MediaIoBaseDownload(fh, request)
    done = False
    while done is False:
        status, done = downloader.next_chunk()
        print("Download %d%%." % int(status.progress() * 100))
    rows = fh.getvalue().decode("utf-8")
    return csv.DictReader(io.StringIO(rows), **csv_options)


if __name__ == "__main__":
    file_id = '1dkJYb-WVVseZIQLAY0jh_o0s7Bas4Gp5'
    options = {'delimiter': ';'}
    iter_d = download_csv_as_json(file_id, csv_options=options)

    # LOADING DETAILS
    table_n = 'people'
    schema_prefix = 'drive'
    schema_source_suffix = 'csv'

    # you authenticate by passing a credential, such as RDBMS host/user/port/pass or gcp service account json.
    gcp_credential_json_file_path = KEY_FILE_LOCATION

    # SCHEMA CREATION
    data_schema = None
    # data_schema_file_path = f"/Users/adrian/PycharmProjects/sv/dlt/examples/schemas/inferred_drive_csv_{file_id}_schema.yml"
    data_schema_file_path = f"examples/schemas/inferred_drive_csv_{file_id}_schema.yml"

    credentials = GCPPipelineCredentials.from_services_file(gcp_credential_json_file_path, schema_prefix)

    pipeline = Pipeline(schema_source_suffix)

    schema: Schema = None
    # uncomment to use already modified schema, but the auto-inferred schema will also work nicely
    # TODO: schema below needs is not yet finished, more exploration needed
    # schema = Pipeline.load_schema_from_file("examples/schemas/discord_schema.yml")
    pipeline.create_pipeline(credentials, schema=schema)
    # the pipeline created a working directory. you can always attach the pipeline back by just providing the dir to
    # Pipeline::restore_pipeline
    print(pipeline.root_path)
    # /private/var/folders/4q/gzmrslrs03j587990jcqt23h0000gn/T/tmp5e17osrg/

    # and extract it
    pipeline.extract(iter_d, table_name=table_n)
    pipeline.flush()

    # save schema
    schema = pipeline.get_default_schema()
    schema_yaml = schema.as_yaml()
    f = open(data_schema_file_path, "a")
    f.write(schema_yaml)
    f.close()



import os
from typing import Optional

from dlt.common import json
from dlt.common.file_storage import FileStorage
from dlt.common.schema import Schema, StoredSchema


class SchemaStorage:

    STORE_SCHEMA_FILE_PATTERN = "%s_schema.json"
    FOLDER_SCHEMA_FILE = "schema.json"

    def __init__(self, schema_storage_root: str, makedirs: bool = False) -> None:
        self.storage = FileStorage(schema_storage_root, makedirs=makedirs)

    def load_store_schema(self, name: str) -> Schema:
        # loads a schema from a store holding many schemas
        schema_file = self._get_file_by_name(name)
        stored_schema: StoredSchema = json.loads(self.storage.load(schema_file))
        return Schema.from_dict(stored_schema)

    def load_folder_schema(self, from_folder: str) -> Schema:
        # loads schema from a folder containing one default schema
        schema_path = self._get_file_in_folder(from_folder)
        stored_schema: StoredSchema = json.loads(self.storage.load(schema_path))
        return Schema.from_dict(stored_schema)

    def save_store_schema(self, schema: Schema) -> str:
        # save a schema to schema store
        dump = json.dumps(schema.to_dict(), indent=2)
        schema_file = self._get_file_by_name(schema.schema_name)
        return self.storage.save(schema_file, dump)

    def remove_store_schema(self, name: str) -> None:
        schema_file = self._get_file_by_name(name)
        self.storage.delete(schema_file)

    def save_folder_schema(self, schema: Schema, in_folder: str) -> str:
        # save a schema to a folder holding one schema
        dump = json.dumps(schema.to_dict())
        schema_file = self._get_file_in_folder(in_folder)
        return self.storage.save(schema_file, dump)

    def has_store_schema(self, name: str) -> bool:
        schema_file = self._get_file_by_name(name)
        return self.storage.has_file(schema_file)

    def _get_file_by_name(self, name: str) -> str:
        return SchemaStorage.STORE_SCHEMA_FILE_PATTERN % name

    def _get_file_in_folder(self, folder: str) -> str:
        return os.path.join(folder, SchemaStorage.FOLDER_SCHEMA_FILE)   # if folder is None else os.path.join(folder, SchemaStorage.SCHEMA_FILE)



from dlt.pipeline import Pipeline, GCPPipelineCredentials

if __name__ == '__main__':
    credentials = GCPPipelineCredentials.from_services_file("_secrets/project1234_service.json", "mainnet_3")
    # credentials = PostgresPipelineCredentials("redshift", "chat_analytics_rasa", "mainnet_2", "loader", "3.73.90.3")

    pipeline = Pipeline("ethereum")
    # restore pipeline from the working directory
    # working directory contains pipeline full state after it was created
    # it is possible to restore pipeline in many concurrently running components ie. several extractors, separate unpacker and loader
    # pipeline.restore_pipeline(credentials, "/tmp/tmp724aveoc/")
    pipeline.restore_pipeline(credentials, "/tmp/tmpswbi3v0o/")

    # the code below will unpack and load anything that was extracted by extractor process(es)

    # do we have anything to unpack
    print(pipeline.list_extracted_loads())

    # just finalize
    pipeline.flush()



from oauth2client.service_account import ServiceAccountCredentials

from dlt.pipeline import Pipeline, GCPPipelineCredentials

from examples.sources.google_sheets import get_source


p = Pipeline("annotations")
credentials = GCPPipelineCredentials.from_services_file('_secrets/project1234_service.json', "load_3")
p.create_pipeline(credentials)

# create google credentials to get data from source
sheets_credentials = ServiceAccountCredentials.from_json_keyfile_name('_secrets/project1234_service.json')

p.extract(
    get_source(sheets_credentials, "11G95oVZjieRhyGqtQMQqlqpxyvWkRXowKE8CtdLtFaU", "2022-05"),
    table_name="model_2022-05"
)
p.flush()
# display inferred schema
print(p.get_default_schema().as_yaml(remove_default_hints=True))



from typing import Type

from dlt.common.typing import StrAny
from dlt.common.configuration.pool_runner_configuration import TPoolType
from dlt.common.dataset_writers import TWriterType

from dlt.common.configuration import (PoolRunnerConfiguration, UnpackingVolumeConfiguration,
                                              LoadingVolumeConfiguration, SchemaVolumeConfiguration,
                                              ProductionLoadingVolumeConfiguration, ProductionUnpackingVolumeConfiguration,
                                              ProductionSchemaVolumeConfiguration,
                                              TPoolType, make_configuration)

from . import __version__


class UnpackerConfiguration(PoolRunnerConfiguration, UnpackingVolumeConfiguration, LoadingVolumeConfiguration, SchemaVolumeConfiguration):
    MAX_EVENTS_IN_CHUNK: int = 40000  # maximum events to be processed in single chunk
    WRITER_TYPE: TWriterType = "jsonl"  # jsonp or insert commands will be generated
    ADD_EVENT_JSON: bool = True  # add event json to "event" table, useful for debugging or recreating tracker
    POOL_TYPE: TPoolType = "process"


class ProductionUnpackerConfiguration(ProductionUnpackingVolumeConfiguration, ProductionLoadingVolumeConfiguration,
                                      ProductionSchemaVolumeConfiguration, UnpackerConfiguration):
    pass


def configuration(initial_values: StrAny = None) -> Type[UnpackerConfiguration]:
    return make_configuration(UnpackerConfiguration, ProductionUnpackerConfiguration, initial_values=initial_values)
